# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris._concatenate.concatenate.py`."""

# import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

import cf_units
import numpy as np
import numpy.ma as ma

from iris._concatenate import concatenate
from iris._lazy_data import as_lazy_data
import iris.coords
import iris.cube
from iris.exceptions import ConcatenateError


class TestEpoch(tests.IrisTest):
    def simple_1d_time_cubes(self, reftimes, coords_points):
        cubes = []
        data_points = [273, 275, 278, 277, 274]
        for reftime, coord_points in zip(reftimes, coords_points):
            cube = iris.cube.Cube(
                np.array(data_points, dtype=np.float32),
                standard_name="air_temperature",
                units="K",
            )
            unit = cf_units.Unit(reftime, calendar="gregorian")
            coord = iris.coords.DimCoord(
                points=np.array(coord_points, dtype=np.float32),
                standard_name="time",
                units=unit,
            )
            cube.add_dim_coord(coord, 0)
            cubes.append(cube)
        return cubes

    def test_concat_1d_with_same_time_units(self):
        reftimes = [
            "hours since 1970-01-01 00:00:00",
            "hours since 1970-01-01 00:00:00",
        ]
        coords_points = [[1, 2, 3, 4, 5], [6, 7, 8, 9, 10]]
        cubes = self.simple_1d_time_cubes(reftimes, coords_points)
        result = concatenate(cubes)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (10,))


class TestMessages(tests.IrisTest):
    def setUp(self):
        data = np.arange(24, dtype=np.float32).reshape(2, 3, 4)
        cube = iris.cube.Cube(data, standard_name="air_temperature", units="K")
        # Time coord
        t_unit = cf_units.Unit(
            "hours since 1970-01-01 00:00:00", calendar="gregorian"
        )
        t_coord = iris.coords.DimCoord(
            points=np.arange(2, dtype=np.float32),
            standard_name="time",
            units=t_unit,
        )
        cube.add_dim_coord(t_coord, 0)
        # Lats and lons
        x_coord = iris.coords.DimCoord(
            points=np.arange(3, dtype=np.float32),
            standard_name="longitude",
            units="degrees",
        )
        cube.add_dim_coord(x_coord, 1)
        y_coord = iris.coords.DimCoord(
            points=np.arange(4, dtype=np.float32),
            standard_name="latitude",
            units="degrees",
        )
        cube.add_dim_coord(y_coord, 2)
        # Scalars
        cube.add_aux_coord(iris.coords.AuxCoord([0], "height", units="m"))
        # Aux Coords
        cube.add_aux_coord(
            iris.coords.AuxCoord(data, long_name="wibble", units="1"),
            data_dims=(0, 1, 2),
        )
        cube.add_aux_coord(
            iris.coords.AuxCoord([0, 1, 2], long_name="foo", units="1"),
            data_dims=(1,),
        )
        cube.add_cell_measure(
            iris.coords.CellMeasure([0, 1, 2], long_name="bar", units="1"),
            data_dims=(1,),
        )
        cube.add_ancillary_variable(
            iris.coords.AncillaryVariable(
                [0, 1, 2], long_name="baz", units="1"
            ),
            data_dims=(1,),
        )
        self.cube = cube

    def test_definition_difference_message(self):
        cube_1 = self.cube
        cube_2 = cube_1.copy()
        cube_2.units = "1"
        exc_regexp = "Cube metadata differs for phenomenon: *"
        with self.assertRaisesRegex(ConcatenateError, exc_regexp):
            _ = concatenate([cube_1, cube_2], True)

    def test_dimensions_difference_message(self):
        cube_1 = self.cube
        cube_2 = cube_1.copy()
        cube_2.remove_coord("latitude")
        exc_regexp = "Dimension coordinates differ: .* != .*"
        with self.assertRaisesRegex(ConcatenateError, exc_regexp):
            _ = concatenate([cube_1, cube_2], True)

    def test_dimensions_metadata_difference_message(self):
        cube_1 = self.cube
        cube_2 = cube_1.copy()
        cube_2.coord("latitude").long_name = "bob"
        exc_regexp = "Dimension coordinates metadata differ: .* != .*"
        with self.assertRaisesRegex(ConcatenateError, exc_regexp):
            _ = concatenate([cube_1, cube_2], True)

    def test_aux_coords_difference_message(self):
        cube_1 = self.cube
        cube_2 = cube_1.copy()
        cube_2.remove_coord("foo")
        exc_regexp = "Auxiliary coordinates differ: .* != .*"
        with self.assertRaisesRegex(ConcatenateError, exc_regexp):
            _ = concatenate([cube_1, cube_2], True)

    def test_aux_coords_metadata_difference_message(self):
        cube_1 = self.cube
        cube_2 = cube_1.copy()
        cube_2.coord("foo").units = "m"
        exc_regexp = "Auxiliary coordinates metadata differ: .* != .*"
        with self.assertRaisesRegex(ConcatenateError, exc_regexp):
            _ = concatenate([cube_1, cube_2], True)

    def test_scalar_coords_difference_message(self):
        cube_1 = self.cube
        cube_2 = cube_1.copy()
        cube_2.remove_coord("height")
        exc_regexp = "Scalar coordinates differ: .* != .*"
        with self.assertRaisesRegex(ConcatenateError, exc_regexp):
            _ = concatenate([cube_1, cube_2], True)

    def test_scalar_coords_metadata_difference_message(self):
        cube_1 = self.cube
        cube_2 = cube_1.copy()
        cube_2.coord("height").long_name = "alice"
        exc_regexp = "Scalar coordinates metadata differ: .* != .*"
        with self.assertRaisesRegex(ConcatenateError, exc_regexp):
            _ = concatenate([cube_1, cube_2], True)

    def test_cell_measure_difference_message(self):
        cube_1 = self.cube
        cube_2 = cube_1.copy()
        cube_2.remove_cell_measure("bar")
        exc_regexp = "Cell measures differ: .* != .*"
        with self.assertRaisesRegex(ConcatenateError, exc_regexp):
            _ = concatenate([cube_1, cube_2], True)

    def test_cell_measure_metadata_difference_message(self):
        cube_1 = self.cube
        cube_2 = cube_1.copy()
        cube_2.cell_measure("bar").units = "m"
        exc_regexp = "Cell measures metadata differ: .* != .*"
        with self.assertRaisesRegex(ConcatenateError, exc_regexp):
            _ = concatenate([cube_1, cube_2], True)

    def test_ancillary_variable_difference_message(self):
        cube_1 = self.cube
        cube_2 = cube_1.copy()
        cube_2.remove_ancillary_variable("baz")
        exc_regexp = "Ancillary variables differ: .* != .*"
        with self.assertRaisesRegex(ConcatenateError, exc_regexp):
            _ = concatenate([cube_1, cube_2], True)

    def test_ancillary_variable_metadata_difference_message(self):
        cube_1 = self.cube
        cube_2 = cube_1.copy()
        cube_2.ancillary_variable("baz").units = "m"
        exc_regexp = "Ancillary variables metadata differ: .* != .*"
        with self.assertRaisesRegex(ConcatenateError, exc_regexp):
            _ = concatenate([cube_1, cube_2], True)

    def test_ndim_difference_message(self):
        cube_1 = self.cube
        cube_2 = iris.cube.Cube(
            np.arange(5, dtype=np.float32),
            standard_name="air_temperature",
            units="K",
        )
        x_coord = iris.coords.DimCoord(
            points=np.arange(5, dtype=np.float32),
            standard_name="longitude",
            units="degrees",
        )
        cube_2.add_dim_coord(x_coord, 0)
        exc_regexp = "Data dimensions differ: [0-9] != [0-9]"
        with self.assertRaisesRegex(ConcatenateError, exc_regexp):
            _ = concatenate([cube_1, cube_2], True)

    def test_datatype_difference_message(self):
        cube_1 = self.cube
        cube_2 = cube_1.copy()
        cube_2.data.dtype = np.float64
        exc_regexp = "Data types differ: .* != .*"
        with self.assertRaisesRegex(ConcatenateError, exc_regexp):
            _ = concatenate([cube_1, cube_2], True)


class TestOrder(tests.IrisTest):
    def _make_cube(self, points, bounds=None):
        nx = 4
        data = np.arange(len(points) * nx).reshape(len(points), nx)
        cube = iris.cube.Cube(data, standard_name="air_temperature", units="K")
        lat = iris.coords.DimCoord(points, "latitude", bounds=bounds)
        lon = iris.coords.DimCoord(np.arange(nx), "longitude")
        cube.add_dim_coord(lat, 0)
        cube.add_dim_coord(lon, 1)
        return cube

    def test_asc_points(self):
        top = self._make_cube([10, 30, 50, 70, 90])
        bottom = self._make_cube([-90, -70, -50, -30, -10])
        result = concatenate([top, bottom])
        self.assertEqual(len(result), 1)

    def test_asc_bounds(self):
        top = self._make_cube([22.5, 67.5], [[0, 45], [45, 90]])
        bottom = self._make_cube([-67.5, -22.5], [[-90, -45], [-45, 0]])
        result = concatenate([top, bottom])
        self.assertEqual(len(result), 1)

    def test_asc_points_with_singleton_ordered(self):
        top = self._make_cube([5])
        bottom = self._make_cube([15, 25])
        result = concatenate([top, bottom])
        self.assertEqual(len(result), 1)

    def test_asc_points_with_singleton_unordered(self):
        top = self._make_cube([25])
        bottom = self._make_cube([5, 15])
        result = concatenate([top, bottom])
        self.assertEqual(len(result), 1)

    def test_asc_bounds_with_singleton_ordered(self):
        top = self._make_cube([5], [[0, 10]])
        bottom = self._make_cube([15, 25], [[10, 20], [20, 30]])
        result = concatenate([top, bottom])
        self.assertEqual(len(result), 1)

    def test_asc_bounds_with_singleton_unordered(self):
        top = self._make_cube([25], [[20, 30]])
        bottom = self._make_cube([5, 15], [[0, 10], [10, 20]])
        result = concatenate([top, bottom])
        self.assertEqual(len(result), 1)

    def test_desc_points(self):
        top = self._make_cube([90, 70, 50, 30, 10])
        bottom = self._make_cube([-10, -30, -50, -70, -90])
        result = concatenate([top, bottom])
        self.assertEqual(len(result), 1)

    def test_desc_bounds(self):
        top = self._make_cube([67.5, 22.5], [[90, 45], [45, 0]])
        bottom = self._make_cube([-22.5, -67.5], [[0, -45], [-45, -90]])
        result = concatenate([top, bottom])
        self.assertEqual(len(result), 1)

    def test_desc_points_with_singleton_ordered(self):
        top = self._make_cube([25])
        bottom = self._make_cube([15, 5])
        result = concatenate([top, bottom])
        self.assertEqual(len(result), 1)

    def test_desc_points_with_singleton_unordered(self):
        top = self._make_cube([5])
        bottom = self._make_cube([25, 15])
        result = concatenate([top, bottom])
        self.assertEqual(len(result), 1)

    def test_desc_bounds_with_singleton_ordered(self):
        top = self._make_cube([25], [[30, 20]])
        bottom = self._make_cube([15, 5], [[20, 10], [10, 0]])
        result = concatenate([top, bottom])
        self.assertEqual(len(result), 1)

    def test_desc_bounds_with_singleton_unordered(self):
        top = self._make_cube([5], [[10, 0]])
        bottom = self._make_cube([25, 15], [[30, 20], [20, 10]])
        result = concatenate([top, bottom])
        self.assertEqual(len(result), 1)

    def test_points_all_singleton(self):
        top = self._make_cube([5])
        bottom = self._make_cube([15])
        result1 = concatenate([top, bottom])
        result2 = concatenate([bottom, top])
        self.assertEqual(len(result1), 1)
        self.assertEqual(len(result2), 1)
        self.assertEqual(result1, result2)

    def test_asc_bounds_all_singleton(self):
        top = self._make_cube([5], [0, 10])
        bottom = self._make_cube([15], [10, 20])
        result1 = concatenate([top, bottom])
        result2 = concatenate([bottom, top])
        self.assertEqual(len(result1), 1)
        self.assertEqual(len(result2), 1)
        self.assertEqual(result1, result2)

    def test_desc_bounds_all_singleton(self):
        top = self._make_cube([5], [10, 0])
        bottom = self._make_cube([15], [20, 10])
        result1 = concatenate([top, bottom])
        result2 = concatenate([bottom, top])
        self.assertEqual(len(result1), 1)
        self.assertEqual(len(result2), 1)
        self.assertEqual(result1, result2)


class TestConcatenate__dask(tests.IrisTest):
    def build_lazy_cube(self, points, bounds=None, nx=4):
        data = np.arange(len(points) * nx).reshape(len(points), nx)
        data = as_lazy_data(data)
        cube = iris.cube.Cube(data, standard_name="air_temperature", units="K")
        lat = iris.coords.DimCoord(points, "latitude", bounds=bounds)
        lon = iris.coords.DimCoord(np.arange(nx), "longitude")
        cube.add_dim_coord(lat, 0)
        cube.add_dim_coord(lon, 1)
        return cube

    def test_lazy_concatenate(self):
        c1 = self.build_lazy_cube([1, 2])
        c2 = self.build_lazy_cube([3, 4, 5])
        (cube,) = concatenate([c1, c2])
        self.assertTrue(cube.has_lazy_data())
        self.assertFalse(ma.isMaskedArray(cube.data))

    def test_lazy_concatenate_masked_array_mixed_deferred(self):
        c1 = self.build_lazy_cube([1, 2])
        c2 = self.build_lazy_cube([3, 4, 5])
        c2.data = np.ma.masked_greater(c2.data, 3)
        (cube,) = concatenate([c1, c2])
        self.assertTrue(cube.has_lazy_data())
        self.assertTrue(ma.isMaskedArray(cube.data))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test area weighted regridding.

"""

# import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

import copy
import random

import dask.array as da
import numpy as np
import numpy.ma as ma

from iris.experimental.regrid import (
    regrid_area_weighted_rectilinear_src_and_grid as regrid_area_weighted,
)
import iris.analysis._interpolation
import iris.tests.stock

RESULT_DIR = (
    "experimental",
    "regrid",
    "regrid_area_weighted_rectilinear_src_and_grid",
)


def _scaled_and_offset_grid(
    cube, x_scalefactor, y_scalefactor, x_offset=0.0, y_offset=0.0
):
    """
    Return a cube with a horizontal grid that is scaled and offset
    from the horizontal grid of `src`.

    """
    x, y = iris.analysis._interpolation.get_xy_dim_coords(cube)
    new_cube = cube.copy()
    new_cube.replace_coord(x * x_scalefactor + x_offset)
    new_cube.replace_coord(y * y_scalefactor + y_offset)
    return new_cube


def _subsampled_coord(coord, subsamplefactor):
    """
    Return a coordinate that is a subsampled copy of `coord`.

    .. note:: `subsamplefactor` must be an integer >= 1.

    """
    if not isinstance(subsamplefactor, int):
        raise ValueError("subsamplefactor must be an integer.")
    if subsamplefactor < 1:
        raise ValueError("subsamplefactor must be >= 1.")
    if not coord.has_bounds():
        raise ValueError("The coordinate must have bounds.")
    new_coord = coord[::subsamplefactor]
    new_bounds = new_coord.bounds.copy()
    new_bounds[:, 1] = coord.bounds[
        (subsamplefactor - 1) :: subsamplefactor, 1
    ]
    new_bounds[-1, 1] = coord.bounds[-1, 1]
    new_coord = coord.copy(points=new_coord.points, bounds=new_bounds)
    return new_coord


def _subsampled_grid(cube, x_subsamplefactor, y_subsamplefactor):
    """
    Return a cube that has a horizontal grid that is a subsampled
    version of the horizontal grid of `cube`.

    .. note:: The two subsamplefactors must both be integers >= 1.

    .. note:: The data of the returned cube is populated with zeros.

    """
    x, y = iris.analysis._interpolation.get_xy_dim_coords(cube)
    x_dim = cube.coord_dims(x)[0]
    y_dim = cube.coord_dims(y)[0]
    new_x = _subsampled_coord(x, x_subsamplefactor)
    new_y = _subsampled_coord(y, y_subsamplefactor)
    new_shape = list(cube.shape)
    new_shape[x_dim] = len(new_x.points)
    new_shape[y_dim] = len(new_y.points)
    new_data = np.zeros(new_shape)
    new_cube = iris.cube.Cube(new_data)
    new_cube.metadata = cube.metadata
    new_cube.add_dim_coord(new_y, y_dim)
    new_cube.add_dim_coord(new_x, x_dim)
    return new_cube


def _resampled_coord(coord, samplefactor):
    """
    Return a coordinate that has the same extent as `coord` but has
    `samplefactor` times as many points and bounds.

    """
    bounds = coord.bounds
    lower = bounds[0, 0]
    upper = bounds[-1, 1]
    # Prevent fp-precision increasing the extent by "squeezing" the grid.
    delta = 0.00001 * np.sign(upper - lower) * abs(bounds[0, 1] - bounds[0, 0])
    lower = lower + delta
    upper = upper - delta
    samples = int(len(bounds) * samplefactor)
    new_points, step = np.linspace(
        lower, upper, samples, endpoint=False, retstep=True
    )
    new_points += step * 0.5
    new_coord = coord.copy(points=new_points)
    new_coord.guess_bounds()
    return new_coord


def _resampled_grid(cube, x_samplefactor, y_samplefactor):
    """
    Return a cube that has the same horizontal extent as `cube` but has
    a reduced (or increased) number of points (and bounds) along the X and Y
    dimensions.

    The resulting number of points for each dimension is determined by::

        int(len(coord.points) * samplefactor)

    This will be truncated if the result is not an integer.

    .. note:: The data of the returned cube is populated with zeros.

    """
    x, y = iris.analysis._interpolation.get_xy_dim_coords(cube)
    x_dim = cube.coord_dims(x)[0]
    y_dim = cube.coord_dims(y)[0]
    new_x = _resampled_coord(x, x_samplefactor)
    new_y = _resampled_coord(y, y_samplefactor)
    new_shape = list(cube.shape)
    new_shape[x_dim] = len(new_x.points)
    new_shape[y_dim] = len(new_y.points)
    new_data = np.zeros(new_shape)
    new_cube = iris.cube.Cube(new_data)
    new_cube.metadata = cube.metadata
    new_cube.add_dim_coord(new_y, y_dim)
    new_cube.add_dim_coord(new_x, x_dim)
    return new_cube


@tests.skip_data
class TestAreaWeightedRegrid(tests.IrisTest):
    def setUp(self):
        # A cube with a hybrid height derived coordinate.
        self.realistic_cube = iris.tests.stock.realistic_4d()[:2, :5, :20, :30]
        # A simple (3, 4) cube.
        self.simple_cube = iris.tests.stock.lat_lon_cube()
        self.simple_cube.coord("latitude").guess_bounds(0.0)
        self.simple_cube.coord("longitude").guess_bounds(0.0)

    def test_no_bounds(self):
        src = self.simple_cube.copy()
        src.coord("latitude").bounds = None
        dest = self.simple_cube.copy()
        with self.assertRaises(ValueError):
            regrid_area_weighted(src, dest)

        src = self.simple_cube.copy()
        src.coord("longitude").bounds = None
        with self.assertRaises(ValueError):
            regrid_area_weighted(src, dest)

        src = self.simple_cube.copy()
        dest = self.simple_cube.copy()
        dest.coord("latitude").bounds = None
        with self.assertRaises(ValueError):
            regrid_area_weighted(src, dest)

        dest = self.simple_cube.copy()
        dest.coord("longitude").bounds = None
        with self.assertRaises(ValueError):
            regrid_area_weighted(src, dest)

    def test_non_contiguous_bounds(self):
        src = self.simple_cube.copy()
        bounds = src.coord("latitude").bounds.copy()
        bounds[1, 1] -= 0.1
        src.coord("latitude").bounds = bounds
        dest = self.simple_cube.copy()
        with self.assertRaises(ValueError):
            regrid_area_weighted(src, dest)

        src = self.simple_cube.copy()
        dest = self.simple_cube.copy()
        bounds = dest.coord("longitude").bounds.copy()
        bounds[1, 1] -= 0.1
        dest.coord("longitude").bounds = bounds
        with self.assertRaises(ValueError):
            regrid_area_weighted(src, dest)

    def test_missing_coords(self):
        dest = self.simple_cube.copy()
        # Missing src_x.
        src = self.simple_cube.copy()
        src.remove_coord("longitude")
        with self.assertRaises(ValueError):
            regrid_area_weighted(src, dest)
        # Missing src_y.
        src = self.simple_cube.copy()
        src.remove_coord("latitude")
        with self.assertRaises(ValueError):
            regrid_area_weighted(src, dest)
        # Missing dest_x.
        src = self.simple_cube.copy()
        dest = self.simple_cube.copy()
        dest.remove_coord("longitude")
        with self.assertRaises(ValueError):
            regrid_area_weighted(src, dest)
        # Missing dest_y.
        src = self.simple_cube.copy()
        dest = self.simple_cube.copy()
        dest.remove_coord("latitude")
        with self.assertRaises(ValueError):
            regrid_area_weighted(src, dest)

    def test_different_cs(self):
        src = self.simple_cube.copy()
        src_cs = copy.copy(src.coord("latitude").coord_system)
        src_cs.semi_major_axis = 7000000
        src.coord("longitude").coord_system = src_cs
        src.coord("latitude").coord_system = src_cs
        dest = self.simple_cube.copy()
        dest_cs = copy.copy(src_cs)
        dest_cs.semi_major_axis = 7000001
        dest.coord("longitude").coord_system = dest_cs
        dest.coord("latitude").coord_system = dest_cs
        with self.assertRaises(ValueError):
            regrid_area_weighted(src, dest)

    def test_regrid_to_same_grid(self):
        src = self.simple_cube
        res = regrid_area_weighted(src, src)
        self.assertEqual(res, src)
        self.assertCMLApproxData(res, RESULT_DIR + ("simple.cml",))

    def test_equal_area_numbers(self):
        # Remove coords system and units so it is no longer spherical.
        self.simple_cube.coord("latitude").coord_system = None
        self.simple_cube.coord("latitude").units = None
        self.simple_cube.coord("longitude").coord_system = None
        self.simple_cube.coord("longitude").units = None
        # Reduce to a single cell
        src = self.simple_cube.copy()
        dest = _subsampled_grid(src, 4, 3)
        res = regrid_area_weighted(src, dest)
        expected_val = np.mean(src.data)
        self.assertAlmostEqual(expected_val, res.data)

        # Reduce to two cells along x
        src = self.simple_cube.copy()
        dest = _subsampled_grid(src, 2, 3)
        res = regrid_area_weighted(src, dest)
        expected_val_left = np.mean(src.data[:, 0:2])
        self.assertEqual(expected_val_left, res.data[0])
        expected_val_right = np.mean(src.data[:, 2:4])
        self.assertAlmostEqual(expected_val_right, res.data[1])

        # Reduce to two cells along x, one three times the size
        # of the other.
        src = self.simple_cube.copy()
        dest = _subsampled_grid(src, 2, 3)
        lon = dest.coord("longitude")
        points = lon.points.copy()
        bounds = [[-1, 0], [0, 3]]
        lon = lon.copy(points=points, bounds=bounds)
        dest.replace_coord(lon)
        res = regrid_area_weighted(src, dest)
        expected_val_left = np.mean(src.data[:, 0:1])
        self.assertEqual(expected_val_left, res.data[0])
        expected_val_right = np.mean(src.data[:, 1:4])
        self.assertAlmostEqual(expected_val_right, res.data[1])

    def test_unqeual_area_numbers(self):
        # Remove coords system and units so it is no longer spherical.
        self.simple_cube.coord("latitude").coord_system = None
        self.simple_cube.coord("latitude").units = None
        self.simple_cube.coord("longitude").coord_system = None
        self.simple_cube.coord("longitude").units = None
        # Reduce src to two cells along x, one three times the size
        # of the other.
        src = self.simple_cube.copy()
        src = _subsampled_grid(src, 2, 2)
        lon = src.coord("longitude")
        points = lon.points.copy()
        bounds = [[-1, 0], [0, 3]]
        lon = lon.copy(points=points, bounds=bounds)
        src.replace_coord(lon)
        # Reduce src to two cells along y, one 2 times the size
        # of the other.
        lat = src.coord("latitude")
        points = lat.points.copy()
        bounds = [[-1, 0], [0, 2]]
        lat = lat.copy(points=points, bounds=bounds)
        src.replace_coord(lat)
        # Populate with data
        src.data = np.arange(src.data.size).reshape(src.shape) + 1.23
        # dest is a single cell over the whole area.
        dest = _subsampled_grid(self.simple_cube, 4, 3)
        res = regrid_area_weighted(src, dest)
        expected_val = (
            1.0 / 12.0 * src.data[0, 0]
            + 2.0 / 12.0 * np.mean(src.data[1:, 0])
            + 3.0 / 12.0 * np.mean(src.data[0, 1:])
            + 6.0 / 12.0 * np.mean(src.data[1:, 1:])
        )
        self.assertAlmostEqual(expected_val, res.data)

    def test_regrid_latlon_reduced_res(self):
        src = self.simple_cube
        # Reduce from (3, 4) to (2, 2).
        dest = _subsampled_grid(src, 2, 2)
        res = regrid_area_weighted(src, dest)
        self.assertCMLApproxData(res, RESULT_DIR + ("latlonreduced.cml",))

    def test_regrid_reorder_axis(self):
        src = self.realistic_cube[0, :4, :3, :2]
        z = src.coord("model_level_number")
        lat = src.coord("grid_latitude")
        lon = src.coord("grid_longitude")
        dest = _resampled_grid(self.realistic_cube[0, 0, :3, :2], 3, 3)
        res = regrid_area_weighted(src, dest)
        self.assertArrayShapeStats(src, (4, 3, 2), 288.08868, 0.008262919)
        self.assertArrayShapeStats(res, (4, 9, 6), 288.08865, 0.00826281)
        # Reshape src so that the coords are ordered [x, z, y],
        # the mean and std statistics should be the same
        data = np.moveaxis(src.data.copy(), 2, 0)
        src = iris.cube.Cube(data)
        src.add_dim_coord(lat, 2)
        src.add_dim_coord(z, 1)
        src.add_dim_coord(lon, 0)
        res = regrid_area_weighted(src, dest)
        self.assertArrayShapeStats(src, (2, 4, 3), 288.08868, 0.008262919)
        self.assertArrayShapeStats(res, (6, 4, 9), 288.08865, 0.00826281)
        # Reshape src so that the coords are ordered [y, x, z],
        # the mean and std statistics should be the same
        data = np.moveaxis(src.data.copy(), 2, 0)
        src = iris.cube.Cube(data)
        src.add_dim_coord(z, 2)
        src.add_dim_coord(lon, 1)
        src.add_dim_coord(lat, 0)
        dest = _resampled_grid(self.realistic_cube[0, 0, :3, :2], 3, 3)
        res = regrid_area_weighted(src, dest)
        self.assertArrayShapeStats(src, (3, 2, 4), 288.08868, 0.008262919)
        self.assertArrayShapeStats(res, (9, 6, 4), 288.08865, 0.00826281)

    def test_regrid_lon_to_half_res(self):
        src = self.simple_cube
        dest = _resampled_grid(src, 0.5, 1.0)
        res = regrid_area_weighted(src, dest)
        self.assertCMLApproxData(res, RESULT_DIR + ("lonhalved.cml",))

    def test_regrid_to_non_int_frac(self):
        # Create dest such that bounds do not line up
        # with src: src.shape = (3, 4), dest.shape = (2, 3)
        src = self.simple_cube
        dest = _resampled_grid(src, 0.75, 0.67)
        res = regrid_area_weighted(src, dest)
        self.assertCMLApproxData(res, RESULT_DIR + ("lower.cml",))

    def test_regrid_to_higher_res(self):
        src = self.simple_cube
        frac = 3.5
        dest = _resampled_grid(src, frac, frac)
        res = regrid_area_weighted(src, dest)
        self.assertCMLApproxData(res, RESULT_DIR + ("higher.cml",))

    def test_hybrid_height(self):
        src = self.realistic_cube
        dest = _resampled_grid(src, 0.7, 0.8)
        res = regrid_area_weighted(src, dest)
        self.assertCMLApproxData(res, RESULT_DIR + ("hybridheight.cml",))

    def test_missing_data(self):
        src = self.simple_cube.copy()
        src.data = ma.masked_array(src.data, fill_value=999)
        src.data[1, 2] = ma.masked
        dest = _resampled_grid(self.simple_cube, 2.3, 2.4)
        res = regrid_area_weighted(src, dest)
        mask = np.zeros((7, 9), bool)
        mask[slice(2, 5), slice(4, 7)] = True
        self.assertArrayEqual(res.data.mask, mask)
        self.assertArrayEqual(res.data.fill_value, 999)

    def test_masked_data_all_false(self):
        src = self.simple_cube.copy()
        src.data = ma.masked_array(src.data, mask=False, fill_value=999)
        dest = _resampled_grid(self.simple_cube, 2.3, 2.4)
        res = regrid_area_weighted(src, dest)
        self.assertArrayEqual(res.data.mask, False)
        self.assertArrayEqual(res.data.fill_value, 999)

    def test_no_x_overlap(self):
        src = self.simple_cube
        dest = _scaled_and_offset_grid(
            src,
            1.0,
            1.0,
            (
                np.max(src.coord("longitude").bounds)
                - np.min(src.coord("longitude").bounds)
            ),
            0.0,
        )
        res = regrid_area_weighted(src, dest)
        self.assertTrue(res.data.mask.all())

    def test_no_y_overlap(self):
        src = self.simple_cube
        dest = _scaled_and_offset_grid(
            src,
            1.0,
            1.0,
            0.0,
            (
                np.max(src.coord("latitude").bounds)
                - np.min(src.coord("latitude").bounds)
            ),
        )
        res = regrid_area_weighted(src, dest)
        self.assertTrue(res.data.mask.all())

    def test_scalar(self):
        src = self.realistic_cube
        i = 2
        j = 3
        dest = src[0, 0, i, j]
        res = regrid_area_weighted(src, dest)
        self.assertEqual(res, src[:, :, i, j])

    def test_one_point(self):
        src = self.simple_cube.copy()
        for n in range(10):
            i = random.randint(0, src.shape[0] - 1)
            j = random.randint(0, src.shape[1] - 1)
            indices = tuple([slice(i, i + 1), slice(j, j + 1)])
            dest = src[indices]
            res = regrid_area_weighted(src, dest)
            self.assertTrue(res, src[indices])

    def test_ten_by_ten_subset(self):
        src = _resampled_grid(self.simple_cube, 20, 20)
        for n in range(10):
            i = random.randint(0, src.shape[0] - 10)
            j = random.randint(0, src.shape[1] - 10)
            indices = tuple([slice(i, i + 10), slice(j, j + 10)])
            dest = src[indices]
            res = regrid_area_weighted(src, dest)
            self.assertEqual(res, src[indices])

    def test_lazy_nop(self):
        src = self.realistic_cube[:2, :3, :10, :10]
        src.data = da.asarray(src.data, chunks=((1, 1), (2, 1), (10,), (10,)))
        res = regrid_area_weighted(src, src)
        self.assertTrue(res.has_lazy_data())
        self.assertEqual(res, src)

    def test_cross_section(self):
        # Slice to get a cross section.
        # Constant latitude
        src = self.realistic_cube[0, :, 10, :]
        lon = _resampled_coord(src.coord("grid_longitude"), 0.6)
        shape = list(src.shape)
        shape[1] = len(lon.points)
        data = np.zeros(shape)
        dest = iris.cube.Cube(data)
        dest.add_dim_coord(lon, 1)
        dest.add_aux_coord(src.coord("grid_latitude").copy(), None)
        res = regrid_area_weighted(src, dest)
        self.assertCMLApproxData(
            res, RESULT_DIR + ("const_lat_cross_section.cml",)
        )
        # Constant latitude, data order [x, z]
        # Using original and transposing the result should give the
        # same answer.
        src.transpose()
        dest.transpose()
        res = regrid_area_weighted(src, dest)
        res.transpose()
        self.assertCMLApproxData(
            res, RESULT_DIR + ("const_lat_cross_section.cml",)
        )

        # Constant longitude
        src = self.realistic_cube[0, :, :, 10]
        lat = _resampled_coord(src.coord("grid_latitude"), 0.6)
        shape = list(src.shape)
        shape[1] = len(lat.points)
        data = np.zeros(shape)
        dest = iris.cube.Cube(data)
        dest.add_dim_coord(lat, 1)
        dest.add_aux_coord(src.coord("grid_longitude").copy(), None)
        res = regrid_area_weighted(src, dest)
        self.assertCMLApproxData(
            res, RESULT_DIR + ("const_lon_cross_section.cml",)
        )
        # Constant longitude, data order [y, z]
        # Using original and transposing the result should give the
        # same answer.
        src.transpose()
        dest.transpose()
        res = regrid_area_weighted(src, dest)
        res.transpose()
        self.assertCMLApproxData(
            res, RESULT_DIR + ("const_lon_cross_section.cml",)
        )

    def test_scalar_source_cube(self):
        src = self.simple_cube[1, 2]
        # Extend dest beyond src grid
        dest = src.copy()
        dest.coord("latitude").bounds = np.array([[-0.5, 1.5]])
        res = regrid_area_weighted(src, dest)
        self.assertTrue(res.data.mask.all())
        # Shrink dest to 1/4 of src
        dest = src.copy()
        dest.coord("latitude").bounds = np.array([[0.25, 0.75]])
        dest.coord("longitude").bounds = np.array([[1.25, 1.75]])
        res = regrid_area_weighted(src, dest)
        self.assertEqual(res.data, src.data)

    @tests.skip_data
    def test_global_data_reduce_res(self):
        src = iris.tests.stock.global_pp()
        src.coord("latitude").guess_bounds()
        src.coord("longitude").guess_bounds()
        dest = _resampled_grid(src, 0.4, 0.3)
        res = regrid_area_weighted(src, dest)
        self.assertArrayShapeStats(res, (21, 38), 280.484932, 15.831545)

    @tests.skip_data
    def test_global_data_increase_res(self):
        src = iris.tests.stock.global_pp()
        src.coord("latitude").guess_bounds()
        src.coord("longitude").guess_bounds()
        dest = _resampled_grid(src, 1.5, 1.5)
        res = regrid_area_weighted(src, dest)
        self.assertArrayShapeStats(res, (109, 144), 280.349625, 16.073397)

    @tests.skip_data
    def test_global_data_same_res(self):
        src = iris.tests.stock.global_pp()
        src.coord("latitude").guess_bounds()
        src.coord("longitude").guess_bounds()
        res = regrid_area_weighted(src, src)
        self.assertArrayShapeStats(res, (73, 96), 279.945160, 16.345842)

    @tests.skip_data
    def test_global_data_subset(self):
        src = iris.tests.stock.global_pp()
        src.coord("latitude").guess_bounds()
        src.coord("longitude").guess_bounds()
        dest_lat = src.coord("latitude")[0:40]
        dest_lon = iris.coords.DimCoord(
            np.linspace(-160, -70, 30),
            standard_name="longitude",
            units="degrees",
            coord_system=dest_lat.coord_system,
        )
        # Note target grid (in -180 to 180) src in 0 to 360
        dest_lon.guess_bounds()
        data = np.zeros((dest_lat.shape[0], dest_lon.shape[0]))
        dest = iris.cube.Cube(data)
        dest.add_dim_coord(dest_lat, 0)
        dest.add_dim_coord(dest_lon, 1)

        res = regrid_area_weighted(src, dest)
        self.assertArrayShapeStats(res, (40, 30), 280.979310, 16.640421)

    @tests.skip_data
    def test_circular_subset(self):
        src = iris.tests.stock.global_pp()
        src.coord("latitude").guess_bounds()
        src.coord("longitude").guess_bounds()
        dest_lat = src.coord("latitude")[0:40]
        dest_lon = iris.coords.DimCoord(
            [-15.0, -10.0, -5.0, 0.0, 5.0, 10.0, 15.0],
            standard_name="longitude",
            units="degrees",
            coord_system=dest_lat.coord_system,
        )
        # Note target grid (in -180 to 180) src in 0 to 360
        dest_lon.guess_bounds()
        data = np.zeros((dest_lat.shape[0], dest_lon.shape[0]))
        dest = iris.cube.Cube(data)
        dest.add_dim_coord(dest_lat, 0)
        dest.add_dim_coord(dest_lon, 1)

        res = regrid_area_weighted(src, dest)
        self.assertArrayShapeStats(res, (40, 7), 285.653960, 15.212710)

    @tests.skip_data
    def test_non_circular_subset(self):
        src = iris.tests.stock.global_pp()
        src.coord("latitude").guess_bounds()
        src.coord("longitude").guess_bounds()
        src.coord("longitude").circular = False
        dest_lat = src.coord("latitude")[0:40]
        dest_lon = iris.coords.DimCoord(
            [-15.0, -10.0, -5.0, 0.0, 5.0, 10.0, 15.0],
            standard_name="longitude",
            units="degrees",
            coord_system=dest_lat.coord_system,
        )
        # Note target grid (in -180 to 180) src in 0 to 360
        dest_lon.guess_bounds()
        data = np.zeros((dest_lat.shape[0], dest_lon.shape[0]))
        dest = iris.cube.Cube(data)
        dest.add_dim_coord(dest_lat, 0)
        dest.add_dim_coord(dest_lon, 1)

        res = regrid_area_weighted(src, dest)
        self.assertArrayShapeStats(res, (40, 7), 285.550814, 15.190245)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris.fileformats._pyke_rules.compiled_krb.\
fc_rules_cf_fc.build_vertical_perspective_coordinate_system`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from unittest import mock

import iris
from iris.coord_systems import VerticalPerspective
from iris.fileformats._pyke_rules.compiled_krb.fc_rules_cf_fc import \
    build_vertical_perspective_coordinate_system


class TestBuildVerticalPerspectiveCoordinateSystem(tests.IrisTest):
    def _test(self, inverse_flattening=False, no_offsets=False):
        """
        Generic test that can check vertical perspective validity with or
        without inverse flattening, and false_east/northing-s.
        """
        test_easting = 100.0
        test_northing = 200.0
        cf_grid_var_kwargs = {
            'spec': [],
            'latitude_of_projection_origin': 1.0,
            'longitude_of_projection_origin': 2.0,
            'perspective_point_height': 2000000.0,
            'false_easting': test_easting,
            'false_northing': test_northing,
            'semi_major_axis': 6377563.396}

        ellipsoid_kwargs = {'semi_major_axis': 6377563.396}
        if inverse_flattening:
            ellipsoid_kwargs['inverse_flattening'] = 299.3249646
        else:
            ellipsoid_kwargs['semi_minor_axis'] = 6356256.909
        cf_grid_var_kwargs.update(ellipsoid_kwargs)

        if no_offsets:
            del cf_grid_var_kwargs['false_easting']
            del cf_grid_var_kwargs['false_northing']
            test_easting = 0
            test_northing = 0

        cf_grid_var = mock.Mock(**cf_grid_var_kwargs)
        ellipsoid = iris.coord_systems.GeogCS(**ellipsoid_kwargs)

        cs = build_vertical_perspective_coordinate_system(None, cf_grid_var)
        expected = VerticalPerspective(
            latitude_of_projection_origin=cf_grid_var.
            latitude_of_projection_origin,
            longitude_of_projection_origin=cf_grid_var.
            longitude_of_projection_origin,
            perspective_point_height=cf_grid_var.perspective_point_height,
            false_easting=test_easting,
            false_northing=test_northing,
            ellipsoid=ellipsoid)

        self.assertEqual(cs, expected)

    def test_valid(self):
        self._test(inverse_flattening=False)

    def test_inverse_flattening(self):
        # Check when inverse_flattening is provided instead of semi_minor_axis.
        self._test(inverse_flattening=True)

    def test_no_offsets(self):
        # Check when false_easting/northing attributes are absent.
        self._test(no_offsets=True)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :func:`iris.sample_data_path` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import os
import os.path
import shutil
import tempfile
from unittest import mock

from iris import sample_data_path


def _temp_file(sample_dir):
    # Return the full path to a new genuine file within our
    # temporary directory.
    sample_handle, sample_path = tempfile.mkstemp(dir=sample_dir)
    os.close(sample_handle)
    return sample_path


@tests.skip_sample_data
class TestIrisSampleData_path(tests.IrisTest):
    def setUp(self):
        self.sample_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.sample_dir)

    def test_path(self):
        with mock.patch("iris_sample_data.path", self.sample_dir):
            import iris_sample_data

            self.assertEqual(iris_sample_data.path, self.sample_dir)

    def test_call(self):
        sample_file = _temp_file(self.sample_dir)
        with mock.patch("iris_sample_data.path", self.sample_dir):
            result = sample_data_path(os.path.basename(sample_file))
            self.assertEqual(result, sample_file)

    def test_file_not_found(self):
        with mock.patch("iris_sample_data.path", self.sample_dir):
            with self.assertRaisesRegex(
                ValueError, "Sample data .* not found"
            ):
                sample_data_path("foo")

    def test_file_absolute(self):
        with mock.patch("iris_sample_data.path", self.sample_dir):
            with self.assertRaisesRegex(ValueError, "Absolute path"):
                sample_data_path(os.path.abspath("foo"))

    def test_glob_ok(self):
        sample_path = _temp_file(self.sample_dir)
        sample_glob = "?" + os.path.basename(sample_path)[1:]
        with mock.patch("iris_sample_data.path", self.sample_dir):
            result = sample_data_path(sample_glob)
            self.assertEqual(
                result, os.path.join(self.sample_dir, sample_glob)
            )

    def test_glob_not_found(self):
        with mock.patch("iris_sample_data.path", self.sample_dir):
            with self.assertRaisesRegex(
                ValueError, "Sample data .* not found"
            ):
                sample_data_path("foo.*")

    def test_glob_absolute(self):
        with mock.patch("iris_sample_data.path", self.sample_dir):
            with self.assertRaisesRegex(ValueError, "Absolute path"):
                sample_data_path(os.path.abspath("foo.*"))


class TestIrisSampleDataMissing(tests.IrisTest):
    def test_no_iris_sample_data(self):
        self.patch("iris.iris_sample_data", None)
        with self.assertRaisesRegex(ImportError, "Please install"):
            sample_data_path("")


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.cube.Cube` class operators."""

# import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

import operator

import dask.array as da
import numpy as np
import numpy.ma as ma

import iris
from iris._lazy_data import as_lazy_data
from iris.coords import DimCoord


class Test_lazy_maths(tests.IrisTest):
    def build_lazy_cube(self, points, dtype=np.float64, bounds=None, nx=10):
        data = np.arange(len(points) * nx, dtype=dtype) + 1  # Just avoid 0.
        data = data.reshape(len(points), nx)
        data = as_lazy_data(data)
        cube = iris.cube.Cube(data, standard_name="air_temperature", units="K")
        lat = DimCoord(points, "latitude", bounds=bounds)
        lon = DimCoord(np.arange(nx), "longitude")
        cube.add_dim_coord(lat, 0)
        cube.add_dim_coord(lon, 1)
        return cube

    def check_common(self, base_cube, result):
        self.assertTrue(base_cube.has_lazy_data())
        self.assertTrue(result.has_lazy_data())
        self.assertIsInstance(result.lazy_data(), da.core.Array)

    def cube_cube_math_op(self, c1, math_op):
        result = math_op(c1, c1)
        self.check_common(c1, result)
        expected = math_op(c1.data, c1.data)
        self.assertArrayAlmostEqual(result.data, expected)

    def cube_scalar_math_op(self, c1, scalar, math_op, commutative=True):
        result = math_op(c1, scalar)
        if commutative:
            self.assertEqual(math_op(c1, scalar), math_op(scalar, c1))
        self.check_common(c1, result)
        expected = math_op(c1.data, scalar)
        self.assertArrayAlmostEqual(result.data, expected)

    def test_add_cubes__float(self):
        c1 = self.build_lazy_cube([1, 2])
        op = operator.add
        self.cube_cube_math_op(c1, op)

    def test_add_scalar__float(self):
        c1 = self.build_lazy_cube([1, 2])
        scalar = 5
        op = operator.add
        self.cube_scalar_math_op(c1, scalar, op)

    def test_mul_cubes__float(self):
        c1 = self.build_lazy_cube([1, 2])
        op = operator.mul
        self.cube_cube_math_op(c1, op)

    def test_mul_scalar__float(self):
        c1 = self.build_lazy_cube([1, 2])
        scalar = 5
        op = operator.mul
        self.cube_scalar_math_op(c1, scalar, op)

    def test_sub_cubes__float(self):
        c1 = self.build_lazy_cube([1, 2])
        op = operator.sub
        self.cube_cube_math_op(c1, op)

    def test_sub_scalar__float(self):
        c1 = self.build_lazy_cube([1, 2])
        scalar = 5
        op = operator.sub
        self.cube_scalar_math_op(c1, scalar, op, commutative=False)

    def test_div_cubes__float(self):
        c1 = self.build_lazy_cube([1, 2])
        op = operator.truediv
        self.cube_cube_math_op(c1, op)

    def test_div_scalar__float(self):
        c1 = self.build_lazy_cube([1, 2])
        scalar = 5
        op = operator.truediv
        self.cube_scalar_math_op(c1, scalar, op, commutative=False)

    def test_add_cubes__int(self):
        c1 = self.build_lazy_cube([1, 2], dtype=np.int64)
        op = operator.add
        self.cube_cube_math_op(c1, op)

    def test_add_scalar__int(self):
        c1 = self.build_lazy_cube([1, 2], dtype=np.int64)
        scalar = 5
        op = operator.add
        self.cube_scalar_math_op(c1, scalar, op)

    def test_mul_cubes__int(self):
        c1 = self.build_lazy_cube([1, 2], dtype=np.int64)
        op = operator.mul
        self.cube_cube_math_op(c1, op)

    def test_mul_scalar__int(self):
        c1 = self.build_lazy_cube([1, 2], dtype=np.int64)
        scalar = 5
        op = operator.mul
        self.cube_scalar_math_op(c1, scalar, op)

    def test_sub_cubes__int(self):
        c1 = self.build_lazy_cube([1, 2], dtype=np.int64)
        op = operator.sub
        self.cube_cube_math_op(c1, op)

    def test_sub_scalar__int(self):
        c1 = self.build_lazy_cube([1, 2], dtype=np.int64)
        scalar = 5
        op = operator.sub
        self.cube_scalar_math_op(c1, scalar, op, commutative=False)

    def test_div_cubes__int(self):
        c1 = self.build_lazy_cube([1, 2], dtype=np.int64)
        op = operator.truediv
        self.cube_cube_math_op(c1, op)

    def test_div_scalar__int(self):
        c1 = self.build_lazy_cube([1, 2], dtype=np.int64)
        scalar = 5
        op = operator.truediv
        self.cube_scalar_math_op(c1, scalar, op, commutative=False)


class Test_lazy_maths__scalar_cube(tests.IrisTest):
    def build_lazy_cube(self, value, dtype=np.float64):
        data = as_lazy_data(np.array(value, dtype=dtype))
        return iris.cube.Cube(data, standard_name="air_temperature", units="K")

    def setUp(self):
        self.c1 = self.build_lazy_cube(3)
        self.c2 = self.build_lazy_cube(4)
        self.c3 = self.build_lazy_cube(3, dtype=np.int64)
        self.c4 = self.build_lazy_cube(4, dtype=np.int64)

    def check_common(self, c1, c2, math_op):
        cube = math_op(c1, c2)
        data = cube.data
        self.assertTrue(isinstance(data, np.ndarray))
        self.assertEqual(data.shape, ())

    def test_add_scalar__int(self):
        c3, c4, op = self.c3, 5, operator.add
        self.check_common(c3, c4, op)

    def test_add_cubes__int(self):
        c3, c4, op = self.c3, self.c4, operator.add
        self.check_common(c3, c4, op)

    def test_mul_scalar__int(self):
        c3, c4, op = self.c3, 5, operator.mul
        self.check_common(c3, c4, op)

    def test_mul_cubes__int(self):
        c3, c4, op = self.c3, self.c4, operator.mul
        self.check_common(c3, c4, op)

    def test_sub_scalar__int(self):
        c3, c4, op = self.c3, 5, operator.sub
        self.check_common(c3, c4, op)

    def test_sub_cubes__int(self):
        c3, c4, op = self.c3, self.c4, operator.sub
        self.check_common(c3, c4, op)

    def test_div_scalar__int(self):
        c3, c4, op = self.c3, 5, operator.truediv
        self.check_common(c3, c4, op)

    def test_div_cubes__int(self):
        c3, c4, op = self.c3, self.c4, operator.truediv
        self.check_common(c3, c4, op)

    def test_add_scalar__float(self):
        c1, c2, op = self.c1, 5, operator.add
        self.check_common(c1, c2, op)

    def test_add_cubes__float(self):
        c1, c2, op = self.c1, self.c2, operator.add
        self.check_common(c1, c2, op)

    def test_mul_scalar__float(self):
        c1, c2, op = self.c1, 5, operator.mul
        self.check_common(c1, c2, op)

    def test_mul_cubes__float(self):
        c1, c2, op = self.c1, self.c2, operator.mul
        self.check_common(c1, c2, op)

    def test_sub_scalar__float(self):
        c1, c2, op = self.c1, 5, operator.sub
        self.check_common(c1, c2, op)

    def test_sub_cubes__float(self):
        c1, c2, op = self.c1, self.c2, operator.sub
        self.check_common(c1, c2, op)

    def test_div_scalar__float(self):
        c1, c2, op = self.c1, 5, operator.truediv
        self.check_common(c1, c2, op)

    def test_div_cubes__float(self):
        c1, c2, op = self.c1, self.c2, operator.truediv
        self.check_common(c1, c2, op)


class Test_lazy_maths__masked_data(tests.IrisTest):
    def build_lazy_cube(self, dtype=np.float64):
        data = ma.array(
            [[1.0, 1.0], [1.0, 100000.0]], mask=[[0, 0], [0, 1]], dtype=dtype
        )
        data = as_lazy_data(data)
        cube = iris.cube.Cube(data, standard_name="air_temperature", units="K")
        lat = DimCoord([-10, 10], "latitude")
        lon = DimCoord([10, 20], "longitude")
        cube.add_dim_coord(lat, 0)
        cube.add_dim_coord(lon, 1)
        return cube

    def test_subtract__float(self):
        cube_a = self.build_lazy_cube()
        cube_b = self.build_lazy_cube()
        cube_c = cube_a - cube_b
        self.assertTrue(ma.isMaskedArray(cube_c.data))

    def test_subtract__int(self):
        cube_a = self.build_lazy_cube(dtype=np.int64)
        cube_b = self.build_lazy_cube(dtype=np.int64)
        cube_c = cube_a - cube_b
        self.assertTrue(ma.isMaskedArray(cube_c.data))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :func:`iris.fileformats.um._fast_load._convert_collation`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import cf_units
import cftime
import numpy as np

from iris.fileformats.um._fast_load import (
    _convert_collation as convert_collation,
)
import iris.aux_factory
import iris.coord_systems
import iris.coords
import iris.fileformats.pp
import iris.fileformats.rules


COORD_SYSTEM = iris.coord_systems.GeogCS(6371229.0)
LATITUDE = iris.coords.DimCoord(
    [15, 0, -15], "latitude", units="degrees", coord_system=COORD_SYSTEM
)
LONGITUDE = iris.coords.DimCoord(
    [0, 20, 40, 60],
    "longitude",
    units="degrees",
    coord_system=COORD_SYSTEM,
    circular=True,
)


class Test(tests.IrisTest):
    def _field(self):
        # Create PP field for X wind on a regular lat-lon grid.
        header = [0] * 64
        # Define the regular lat-lon grid.
        header[15] = 1  # LBCODE
        header[17] = 3  # LBROW
        header[18] = 4  # LBNPT
        header[55] = 90  # BPLAT
        header[58] = 30  # BZY
        header[59] = -15  # BDY
        header[60] = -20  # BZX
        header[61] = 20  # BDX
        # Define the STASH code m01s00i002.
        header[41] = 2  # LBUSER(4)
        header[44] = 1  # LBUSER(7)
        field = iris.fileformats.pp.PPField3(header)
        return field

    def _check_phenomenon(self, metadata, factory=None):
        if factory is None:
            self.assertEqual(metadata.factories, [])
        else:
            self.assertEqual(metadata.factories, [factory])
        self.assertEqual(metadata.references, [])
        self.assertEqual(metadata.standard_name, "x_wind")
        self.assertIsNone(metadata.long_name)
        self.assertEqual(metadata.units, cf_units.Unit("m s-1"))
        self.assertEqual(metadata.attributes, {"STASH": (1, 0, 2)})
        self.assertEqual(metadata.cell_methods, [])

    def test_all_scalar(self):
        field = self._field()
        field.lbtim = 11
        field.t1 = cftime.datetime(1970, 1, 1, 18)
        field.t2 = cftime.datetime(1970, 1, 1, 12)
        collation = mock.Mock(
            fields=[field], vector_dims_shape=(), element_arrays_and_dims={}
        )
        metadata = convert_collation(collation)
        self._check_phenomenon(metadata)
        coords_and_dims = [(LONGITUDE, 1), (LATITUDE, 0)]
        self.assertEqual(metadata.dim_coords_and_dims, coords_and_dims)
        coords_and_dims = [
            (
                iris.coords.DimCoord(18, "time", units="hours since epoch"),
                None,
            ),
            (
                iris.coords.DimCoord(
                    12, "forecast_reference_time", units="hours since epoch"
                ),
                None,
            ),
            (iris.coords.DimCoord(6, "forecast_period", units="hours"), None),
        ]
        self.assertEqual(metadata.aux_coords_and_dims, coords_and_dims)

    def test_vector_t1(self):
        field = self._field()
        field.lbtim = 11
        field.t2 = cftime.datetime(1970, 1, 1, 12)
        t1 = (
            [
                cftime.datetime(1970, 1, 1, 18),
                cftime.datetime(1970, 1, 2, 0),
                cftime.datetime(1970, 1, 2, 6),
            ],
            [0],
        )
        collation = mock.Mock(
            fields=[field],
            vector_dims_shape=(3,),
            element_arrays_and_dims={"t1": t1},
        )
        metadata = convert_collation(collation)
        self._check_phenomenon(metadata)
        coords_and_dims = [
            (LONGITUDE, 2),
            (LATITUDE, 1),
            (
                iris.coords.DimCoord(
                    [18, 24, 30], "time", units="hours since epoch"
                ),
                (0,),
            ),
        ]
        self.assertEqual(metadata.dim_coords_and_dims, coords_and_dims)
        coords_and_dims = [
            (
                iris.coords.DimCoord(
                    12, "forecast_reference_time", units="hours since epoch"
                ),
                None,
            ),
            (
                iris.coords.DimCoord(
                    [6, 12, 18], "forecast_period", units="hours"
                ),
                (0,),
            ),
        ]
        self.assertEqual(metadata.aux_coords_and_dims, coords_and_dims)

    def test_vector_t2(self):
        field = self._field()
        field.lbtim = 11
        field.t1 = cftime.datetime(1970, 1, 1, 18)
        t2 = (
            [
                cftime.datetime(1970, 1, 1, 12),
                cftime.datetime(1970, 1, 1, 15),
                cftime.datetime(1970, 1, 1, 18),
            ],
            [0],
        )
        collation = mock.Mock(
            fields=[field],
            vector_dims_shape=(3,),
            element_arrays_and_dims={"t2": t2},
        )
        metadata = convert_collation(collation)
        self._check_phenomenon(metadata)
        coords_and_dims = [
            (LONGITUDE, 2),
            (LATITUDE, 1),
            (
                iris.coords.DimCoord(
                    [12, 15, 18],
                    "forecast_reference_time",
                    units="hours since epoch",
                ),
                (0,),
            ),
        ]
        self.assertEqual(metadata.dim_coords_and_dims, coords_and_dims)
        coords_and_dims = [
            (
                iris.coords.DimCoord(18, "time", units="hours since epoch"),
                None,
            ),
            (
                iris.coords.DimCoord(
                    [6, 3, 0.0], "forecast_period", units="hours"
                ),
                (0,),
            ),
        ]
        self.assertEqual(metadata.aux_coords_and_dims, coords_and_dims)

    def test_vector_lbft(self):
        field = self._field()
        field.lbtim = 21
        field.t1 = cftime.datetime(1970, 1, 1, 12)
        field.t2 = cftime.datetime(1970, 1, 1, 18)
        lbft = ([18, 15, 12], [0])
        collation = mock.Mock(
            fields=[field],
            vector_dims_shape=(3,),
            element_arrays_and_dims={"lbft": lbft},
        )
        metadata = convert_collation(collation)
        self._check_phenomenon(metadata)
        coords_and_dims = [
            (LONGITUDE, 2),
            (LATITUDE, 1),
            (
                iris.coords.DimCoord(
                    [0, 3, 6],
                    "forecast_reference_time",
                    units="hours since epoch",
                ),
                (0,),
            ),
        ]
        coords_and_dims = [
            (
                iris.coords.DimCoord(
                    15, "time", units="hours since epoch", bounds=[[12, 18]]
                ),
                None,
            ),
            (
                iris.coords.DimCoord(
                    [15, 12, 9],
                    "forecast_period",
                    units="hours",
                    bounds=[[12, 18], [9, 15], [6, 12]],
                ),
                (0,),
            ),
        ]
        self.assertEqual(metadata.aux_coords_and_dims, coords_and_dims)

    def test_vector_t1_and_t2(self):
        field = self._field()
        field.lbtim = 11
        t1 = (
            [
                cftime.datetime(1970, 1, 2, 6),
                cftime.datetime(1970, 1, 2, 9),
                cftime.datetime(1970, 1, 2, 12),
            ],
            [1],
        )
        t2 = (
            [cftime.datetime(1970, 1, 1, 12), cftime.datetime(1970, 1, 2, 0)],
            [0],
        )
        collation = mock.Mock(
            fields=[field],
            vector_dims_shape=(2, 3),
            element_arrays_and_dims={"t1": t1, "t2": t2},
        )
        metadata = convert_collation(collation)
        self._check_phenomenon(metadata)
        coords_and_dims = [
            (LONGITUDE, 3),
            (LATITUDE, 2),
            (
                iris.coords.DimCoord(
                    [30, 33, 36], "time", units="hours since epoch"
                ),
                (1,),
            ),
            (
                iris.coords.DimCoord(
                    [12, 24],
                    "forecast_reference_time",
                    units="hours since epoch",
                ),
                (0,),
            ),
        ]
        self.assertEqual(metadata.dim_coords_and_dims, coords_and_dims)
        coords_and_dims = [
            (
                iris.coords.AuxCoord(
                    [[18, 21, 24], [6, 9, 12]],
                    "forecast_period",
                    units="hours",
                ),
                (0, 1),
            )
        ]
        self.assertEqual(metadata.aux_coords_and_dims, coords_and_dims)

    def test_vertical_pressure(self):
        field = self._field()
        field.lbvc = 8
        blev = ([1000, 850, 700], (0,))
        lblev = ([1000, 850, 700], (0,))
        collation = mock.Mock(
            fields=[field],
            vector_dims_shape=(3,),
            element_arrays_and_dims={"blev": blev, "lblev": lblev},
        )
        metadata = convert_collation(collation)
        self._check_phenomenon(metadata)
        coords_and_dims = [
            (LONGITUDE, 2),
            (LATITUDE, 1),
            (
                iris.coords.DimCoord(
                    [1000, 850, 700], long_name="pressure", units="hPa"
                ),
                (0,),
            ),
        ]
        self.assertEqual(metadata.dim_coords_and_dims, coords_and_dims)
        coords_and_dims = []
        self.assertEqual(metadata.aux_coords_and_dims, coords_and_dims)

    def test_soil_level(self):
        field = self._field()
        field.lbvc = 6
        points = [10, 20, 30]
        lower = [0] * 3
        upper = [0] * 3
        lblev = (points, (0,))
        brsvd1 = (lower, (0,))
        brlev = (upper, (0,))
        collation = mock.Mock(
            fields=[field],
            vector_dims_shape=(3,),
            element_arrays_and_dims={
                "lblev": lblev,
                "brsvd1": brsvd1,
                "brlev": brlev,
            },
        )
        metadata = convert_collation(collation)
        self._check_phenomenon(metadata)
        level = iris.coords.DimCoord(
            points,
            long_name="soil_model_level_number",
            attributes={"positive": "down"},
            units="1",
        )
        coords_and_dims = [(LONGITUDE, 2), (LATITUDE, 1), (level, (0,))]
        self.assertEqual(metadata.dim_coords_and_dims, coords_and_dims)
        coords_and_dims = []
        self.assertEqual(metadata.aux_coords_and_dims, coords_and_dims)

    def test_soil_depth(self):
        field = self._field()
        field.lbvc = 6
        points = [10, 20, 30]
        lower = [0, 15, 25]
        upper = [15, 25, 35]
        blev = (points, (0,))
        brsvd1 = (lower, (0,))
        brlev = (upper, (0,))
        collation = mock.Mock(
            fields=[field],
            vector_dims_shape=(3,),
            element_arrays_and_dims={
                "blev": blev,
                "brsvd1": brsvd1,
                "brlev": brlev,
            },
        )
        metadata = convert_collation(collation)
        self._check_phenomenon(metadata)
        depth = iris.coords.DimCoord(
            points,
            standard_name="depth",
            bounds=np.vstack((lower, upper)).T,
            units="m",
            attributes={"positive": "down"},
        )
        coords_and_dims = [(LONGITUDE, 2), (LATITUDE, 1), (depth, (0,))]
        self.assertEqual(metadata.dim_coords_and_dims, coords_and_dims)
        coords_and_dims = []
        self.assertEqual(metadata.aux_coords_and_dims, coords_and_dims)

    def test_vertical_hybrid_height(self):
        field = self._field()
        field.lbvc = 65
        blev = ([5, 18, 38], (0,))
        lblev = ([1000, 850, 700], (0,))
        brsvd1 = ([10, 26, 50], (0,))
        brsvd2 = ([0.9989, 0.9970, 0.9944], (0,))
        brlev = ([0, 10, 26], (0,))
        bhrlev = ([1, 0.9989, 0.9970], (0,))
        lblev = ([1, 2, 3], (0,))
        bhlev = ([0.9994, 0.9979, 0.9957], (0,))
        collation = mock.Mock(
            fields=[field],
            vector_dims_shape=(3,),
            element_arrays_and_dims={
                "blev": blev,
                "lblev": lblev,
                "brsvd1": brsvd1,
                "brsvd2": brsvd2,
                "brlev": brlev,
                "bhrlev": bhrlev,
                "lblev": lblev,
                "bhlev": bhlev,
            },
        )
        metadata = convert_collation(collation)
        factory = iris.fileformats.rules.Factory(
            iris.aux_factory.HybridHeightFactory,
            [
                {"long_name": "level_height"},
                {"long_name": "sigma"},
                iris.fileformats.rules.Reference("orography"),
            ],
        )
        self._check_phenomenon(metadata, factory)
        coords_and_dims = [
            (LONGITUDE, 2),
            (LATITUDE, 1),
            (
                iris.coords.DimCoord(
                    [1, 2, 3],
                    "model_level_number",
                    attributes={"positive": "up"},
                    units="1",
                ),
                (0,),
            ),
        ]
        self.assertEqual(metadata.dim_coords_and_dims, coords_and_dims)
        coords_and_dims = [
            (
                iris.coords.DimCoord(
                    [5, 18, 38],
                    long_name="level_height",
                    units="m",
                    bounds=[[0, 10], [10, 26], [26, 50]],
                    attributes={"positive": "up"},
                ),
                (0,),
            ),
            (
                iris.coords.AuxCoord(
                    [0.9994, 0.9979, 0.9957],
                    long_name="sigma",
                    bounds=[[1, 0.9989], [0.9989, 0.9970], [0.9970, 0.9944]],
                    units="1",
                ),
                (0,),
            ),
        ]
        self.assertEqual(metadata.aux_coords_and_dims, coords_and_dims)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :mod:`iris.plot` module."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris.plot import _broadcast_2d as broadcast
from iris.coords import AuxCoord
from iris.tests.stock import simple_2d, lat_lon_cube


@tests.skip_plot
class TestGraphicStringCoord(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        self.cube = simple_2d(with_bounds=True)
        self.cube.add_aux_coord(
            AuxCoord(list("abcd"), long_name="str_coord"), 1
        )
        self.lat_lon_cube = lat_lon_cube()

    def tick_loc_and_label(self, axis_name, axes=None):
        # Intentional lazy import so that subclasses can have an opportunity
        # to change the backend.
        import matplotlib.pyplot as plt

        # Draw the plot to 'fix' the ticks.
        if axes:
            axes.figure.canvas.draw()
        else:
            axes = plt.gca()
            plt.draw()
        axis = getattr(axes, axis_name)

        locations = axis.get_majorticklocs()
        labels = [tick.get_text() for tick in axis.get_ticklabels()]
        return list(zip(locations, labels))

    def assertBoundsTickLabels(self, axis, axes=None):
        actual = self.tick_loc_and_label(axis, axes)
        expected = [
            (-1.0, ""),
            (0.0, "a"),
            (1.0, "b"),
            (2.0, "c"),
            (3.0, "d"),
            (4.0, ""),
        ]
        self.assertEqual(expected, actual)

    def assertPointsTickLabels(self, axis, axes=None):
        actual = self.tick_loc_and_label(axis, axes)
        expected = [(0.0, "a"), (1.0, "b"), (2.0, "c"), (3.0, "d")]
        self.assertEqual(expected, actual)


@tests.skip_plot
class MixinCoords:
    """
    Mixin class of common plotting tests providing 2-dimensional
    permutations of coordinates and anonymous dimensions.

    """

    def _check(self, u, v, data=None):
        self.assertEqual(self.mpl_patch.call_count, 1)
        if data is not None:
            (actual_u, actual_v, actual_data), _ = self.mpl_patch.call_args
            self.assertArrayEqual(actual_data, data)
        else:
            (actual_u, actual_v), _ = self.mpl_patch.call_args
        self.assertArrayEqual(actual_u, u)
        self.assertArrayEqual(actual_v, v)

    def test_foo_bar(self):
        self.draw_func(self.cube, coords=("foo", "bar"))
        u, v = broadcast(self.foo, self.bar)
        self._check(u, v, self.data)

    def test_bar_foo(self):
        self.draw_func(self.cube, coords=("bar", "foo"))
        u, v = broadcast(self.bar, self.foo)
        self._check(u, v, self.dataT)

    def test_foo_0(self):
        self.draw_func(self.cube, coords=("foo", 0))
        u, v = broadcast(self.foo, self.bar_index)
        self._check(u, v, self.data)

    def test_1_bar(self):
        self.draw_func(self.cube, coords=(1, "bar"))
        u, v = broadcast(self.foo_index, self.bar)
        self._check(u, v, self.data)

    def test_1_0(self):
        self.draw_func(self.cube, coords=(1, 0))
        u, v = broadcast(self.foo_index, self.bar_index)
        self._check(u, v, self.data)

    def test_0_foo(self):
        self.draw_func(self.cube, coords=(0, "foo"))
        u, v = broadcast(self.bar_index, self.foo)
        self._check(u, v, self.dataT)

    def test_bar_1(self):
        self.draw_func(self.cube, coords=("bar", 1))
        u, v = broadcast(self.bar, self.foo_index)
        self._check(u, v, self.dataT)

    def test_0_1(self):
        self.draw_func(self.cube, coords=(0, 1))
        u, v = broadcast(self.bar_index, self.foo_index)
        self._check(u, v, self.dataT)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the function
:func:`iris.analysis.cartography.rotate_winds`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np
import numpy.ma as ma

import cartopy.crs as ccrs
from iris.analysis.cartography import rotate_winds, unrotate_pole
from iris.cube import Cube
from iris.coords import DimCoord, AuxCoord
import iris.coord_systems


def uv_cubes(x=None, y=None):
    """Return u, v cubes with a grid in a rotated pole CRS."""
    cs = iris.coord_systems.RotatedGeogCS(
        grid_north_pole_latitude=37.5, grid_north_pole_longitude=177.5
    )
    if x is None:
        x = np.linspace(311.9, 391.1, 6)
    if y is None:
        y = np.linspace(-23.6, 24.8, 5)

    x2d, y2d = np.meshgrid(x, y)
    u = 10 * (2 * np.cos(2 * np.deg2rad(x2d) + 3 * np.deg2rad(y2d + 30)) ** 2)
    v = 20 * np.cos(6 * np.deg2rad(x2d))
    lon = DimCoord(
        x, standard_name="grid_longitude", units="degrees", coord_system=cs
    )
    lat = DimCoord(
        y, standard_name="grid_latitude", units="degrees", coord_system=cs
    )
    u_cube = Cube(u, standard_name="x_wind", units="m/s")
    v_cube = Cube(v, standard_name="y_wind", units="m/s")
    for cube in (u_cube, v_cube):
        cube.add_dim_coord(lat.copy(), 0)
        cube.add_dim_coord(lon.copy(), 1)
    return u_cube, v_cube


def uv_cubes_3d(ref_cube, n_realization=3):
    """
    Return 3d u, v cubes with a grid in a rotated pole CRS taken from
    the provided 2d cube, by adding a realization dimension
    coordinate bound to teh zeroth dimension.

    """
    lat = ref_cube.coord("grid_latitude")
    lon = ref_cube.coord("grid_longitude")
    x2d, y2d = np.meshgrid(lon.points, lat.points)
    u = 10 * (2 * np.cos(2 * np.deg2rad(x2d) + 3 * np.deg2rad(y2d + 30)) ** 2)
    v = 20 * np.cos(6 * np.deg2rad(x2d))
    # Multiply slices by factor to give variation over 0th dim.
    factor = np.arange(1, n_realization + 1).reshape(n_realization, 1, 1)
    u = factor * u
    v = factor * v
    realization = DimCoord(np.arange(n_realization), "realization")
    u_cube = Cube(u, standard_name="x_wind", units="m/s")
    v_cube = Cube(v, standard_name="y_wind", units="m/s")
    for cube in (u_cube, v_cube):
        cube.add_dim_coord(realization.copy(), 0)
        cube.add_dim_coord(lat.copy(), 1)
        cube.add_dim_coord(lon.copy(), 2)
    return u_cube, v_cube


class TestPrerequisites(tests.IrisTest):
    def test_different_coord_systems(self):
        u, v = uv_cubes()
        v.coord("grid_latitude").coord_system = iris.coord_systems.GeogCS(1)
        with self.assertRaisesRegex(
            ValueError, "Coordinates differ between u and v cubes"
        ):
            rotate_winds(u, v, iris.coord_systems.OSGB())

    def test_different_xy_coord_systems(self):
        u, v = uv_cubes()
        u.coord("grid_latitude").coord_system = iris.coord_systems.GeogCS(1)
        v.coord("grid_latitude").coord_system = iris.coord_systems.GeogCS(1)
        with self.assertRaisesRegex(
            ValueError, "Coordinate systems of x and y coordinates differ"
        ):
            rotate_winds(u, v, iris.coord_systems.OSGB())

    def test_different_shape(self):
        x = np.linspace(311.9, 391.1, 6)
        y = np.linspace(-23.6, 24.8, 5)
        u, _ = uv_cubes(x, y)
        _, v = uv_cubes(x[:-1], y)
        with self.assertRaisesRegex(ValueError, "same shape"):
            rotate_winds(u, v, iris.coord_systems.OSGB())

    def test_xy_dimensionality(self):
        u, v = uv_cubes()
        # Replace 1d lat with 2d lat.
        x = u.coord("grid_longitude").points
        y = u.coord("grid_latitude").points
        x2d, y2d = np.meshgrid(x, y)
        lat_2d = AuxCoord(
            y2d,
            "grid_latitude",
            units="degrees",
            coord_system=u.coord("grid_latitude").coord_system,
        )
        for cube in (u, v):
            cube.remove_coord("grid_latitude")
            cube.add_aux_coord(lat_2d.copy(), (0, 1))

        with self.assertRaisesRegex(
            ValueError,
            "x and y coordinates must have the same number of dimensions",
        ):
            rotate_winds(u, v, iris.coord_systems.OSGB())

    def test_dim_mapping(self):
        x = np.linspace(311.9, 391.1, 3)
        y = np.linspace(-23.6, 24.8, 3)
        u, v = uv_cubes(x, y)
        v.transpose()
        with self.assertRaisesRegex(ValueError, "Dimension mapping"):
            rotate_winds(u, v, iris.coord_systems.OSGB())


class TestAnalyticComparison(tests.IrisTest):
    @staticmethod
    def _unrotate_equation(
        rotated_lons, rotated_lats, rotated_us, rotated_vs, pole_lon, pole_lat
    ):
        # Perform a rotated-pole 'unrotate winds' transformation on arrays of
        # rotated-lat, rotated-lon, u and v.
        # This can be defined as an analytic function : cf. UMDP015

        # Work out the rotation angles.
        lambda_angle = np.radians(pole_lon - 180.0)
        phi_angle = np.radians(90.0 - pole_lat)

        # Get the locations in true lats+lons.
        trueLongitude, trueLatitude = unrotate_pole(
            rotated_lons, rotated_lats, pole_lon, pole_lat
        )

        # Calculate inter-coordinate rotation coefficients.
        cos_rot = np.cos(np.radians(rotated_lons)) * np.cos(
            np.radians(trueLongitude) - lambda_angle
        ) + np.sin(np.radians(rotated_lons)) * np.sin(
            np.radians(trueLongitude) - lambda_angle
        ) * np.cos(
            phi_angle
        )
        sin_rot = -(
            (
                np.sin(np.radians(trueLongitude) - lambda_angle)
                * np.sin(phi_angle)
            )
            / np.cos(np.radians(rotated_lats))
        )

        # Matrix-multiply to rotate the vectors.
        u_true = rotated_us * cos_rot - rotated_vs * sin_rot
        v_true = rotated_vs * cos_rot + rotated_us * sin_rot

        return u_true, v_true

    def _check_rotated_to_true(self, u_rot, v_rot, target_cs, **kwds):
        # Run test calculation (numeric).
        u_true, v_true = rotate_winds(u_rot, v_rot, target_cs)

        # Perform same calculation via the reference method (equations).
        cs_rot = u_rot.coord("grid_longitude").coord_system
        pole_lat = cs_rot.grid_north_pole_latitude
        pole_lon = cs_rot.grid_north_pole_longitude
        rotated_lons = u_rot.coord("grid_longitude").points
        rotated_lats = u_rot.coord("grid_latitude").points
        rotated_lons_2d, rotated_lats_2d = np.meshgrid(
            rotated_lons, rotated_lats
        )
        rotated_u, rotated_v = u_rot.data, v_rot.data
        u_ref, v_ref = self._unrotate_equation(
            rotated_lons_2d,
            rotated_lats_2d,
            rotated_u,
            rotated_v,
            pole_lon,
            pole_lat,
        )

        # Check that all the numerical results are within given tolerances.
        self.assertArrayAllClose(u_true.data, u_ref, **kwds)
        self.assertArrayAllClose(v_true.data, v_ref, **kwds)

    def test_rotated_to_true__small(self):
        # Check for a small field with varying data.
        target_cs = iris.coord_systems.GeogCS(6371229)
        u_rot, v_rot = uv_cubes()
        self._check_rotated_to_true(
            u_rot, v_rot, target_cs, rtol=1e-5, atol=0.0005
        )

    def test_rotated_to_true_global(self):
        # Check for global fields with various constant wind values
        # - constant in the rotated pole system, that is.
        # We expect less accuracy where this gets close to the true poles.
        target_cs = iris.coord_systems.GeogCS(6371229)
        u_rot, v_rot = uv_cubes(
            x=np.arange(0, 360.0, 15), y=np.arange(-89, 89, 10)
        )
        for vector in ((1, 0), (0, 1), (1, 1), (-3, -1.5)):
            u_rot.data[...] = vector[0]
            v_rot.data[...] = vector[1]
            self._check_rotated_to_true(
                u_rot,
                v_rot,
                target_cs,
                rtol=5e-4,
                atol=5e-4,
                err_msg="vector={}".format(vector),
            )


class TestRotatedToOSGB(tests.IrisTest):
    # Define some coordinate ranges for the uv_cubes 'standard' RotatedPole
    # system, that exceed the OSGB margins, but not by "too much".
    _rp_x_min, _rp_x_max = -5.0, 5.0
    _rp_y_min, _rp_y_max = -5.0, 15.0

    def _uv_cubes_limited_extent(self):
        # Make test cubes suitable for transforming to OSGB, as the standard
        # 'uv_cubes' result goes too far outside, leading to errors.
        x = np.linspace(self._rp_x_min, self._rp_x_max, 6)
        y = np.linspace(self._rp_y_min, self._rp_y_max, 5)
        return uv_cubes(x=x, y=y)

    def test_name(self):
        u, v = self._uv_cubes_limited_extent()
        u.rename("bob")
        v.rename("alice")
        ut, vt = rotate_winds(u, v, iris.coord_systems.OSGB())
        self.assertEqual(ut.name(), "transformed_" + u.name())
        self.assertEqual(vt.name(), "transformed_" + v.name())

    def test_new_coords(self):
        u, v = self._uv_cubes_limited_extent()
        x = u.coord("grid_longitude").points
        y = u.coord("grid_latitude").points
        x2d, y2d = np.meshgrid(x, y)
        src_crs = ccrs.RotatedPole(pole_longitude=177.5, pole_latitude=37.5)
        tgt_crs = ccrs.OSGB()
        xyz_tran = tgt_crs.transform_points(src_crs, x2d, y2d)

        ut, vt = rotate_winds(u, v, iris.coord_systems.OSGB())

        points = xyz_tran[..., 0].reshape(x2d.shape)
        expected_x = AuxCoord(
            points,
            standard_name="projection_x_coordinate",
            units="m",
            coord_system=iris.coord_systems.OSGB(),
        )
        self.assertEqual(ut.coord("projection_x_coordinate"), expected_x)
        self.assertEqual(vt.coord("projection_x_coordinate"), expected_x)

        points = xyz_tran[..., 1].reshape(y2d.shape)
        expected_y = AuxCoord(
            points,
            standard_name="projection_y_coordinate",
            units="m",
            coord_system=iris.coord_systems.OSGB(),
        )
        self.assertEqual(ut.coord("projection_y_coordinate"), expected_y)
        self.assertEqual(vt.coord("projection_y_coordinate"), expected_y)

    def test_new_coords_transposed(self):
        u, v = self._uv_cubes_limited_extent()
        # Transpose cubes so that cube is in xy order rather than the
        # typical yx order of meshgrid.
        u.transpose()
        v.transpose()
        x = u.coord("grid_longitude").points
        y = u.coord("grid_latitude").points
        x2d, y2d = np.meshgrid(x, y)
        src_crs = ccrs.RotatedPole(pole_longitude=177.5, pole_latitude=37.5)
        tgt_crs = ccrs.OSGB()
        xyz_tran = tgt_crs.transform_points(src_crs, x2d, y2d)

        ut, vt = rotate_winds(u, v, iris.coord_systems.OSGB())

        points = xyz_tran[..., 0].reshape(x2d.shape)
        expected_x = AuxCoord(
            points,
            standard_name="projection_x_coordinate",
            units="m",
            coord_system=iris.coord_systems.OSGB(),
        )
        self.assertEqual(ut.coord("projection_x_coordinate"), expected_x)
        self.assertEqual(vt.coord("projection_x_coordinate"), expected_x)

        points = xyz_tran[..., 1].reshape(y2d.shape)
        expected_y = AuxCoord(
            points,
            standard_name="projection_y_coordinate",
            units="m",
            coord_system=iris.coord_systems.OSGB(),
        )
        self.assertEqual(ut.coord("projection_y_coordinate"), expected_y)
        self.assertEqual(vt.coord("projection_y_coordinate"), expected_y)
        # Check dim mapping for 2d coords is yx.
        expected_dims = u.coord_dims("grid_latitude") + u.coord_dims(
            "grid_longitude"
        )
        self.assertEqual(
            ut.coord_dims("projection_x_coordinate"), expected_dims
        )
        self.assertEqual(
            ut.coord_dims("projection_y_coordinate"), expected_dims
        )
        self.assertEqual(
            vt.coord_dims("projection_x_coordinate"), expected_dims
        )
        self.assertEqual(
            vt.coord_dims("projection_y_coordinate"), expected_dims
        )

    def test_orig_coords(self):
        u, v = self._uv_cubes_limited_extent()
        ut, vt = rotate_winds(u, v, iris.coord_systems.OSGB())
        self.assertEqual(u.coord("grid_latitude"), ut.coord("grid_latitude"))
        self.assertEqual(v.coord("grid_latitude"), vt.coord("grid_latitude"))
        self.assertEqual(u.coord("grid_longitude"), ut.coord("grid_longitude"))
        self.assertEqual(v.coord("grid_longitude"), vt.coord("grid_longitude"))

    def test_magnitude_preservation(self):
        u, v = self._uv_cubes_limited_extent()
        ut, vt = rotate_winds(u, v, iris.coord_systems.OSGB())
        orig_sq_mag = u.data ** 2 + v.data ** 2
        res_sq_mag = ut.data ** 2 + vt.data ** 2
        self.assertArrayAllClose(orig_sq_mag, res_sq_mag, rtol=5e-4)

    def test_data_values(self):
        u, v = self._uv_cubes_limited_extent()
        # Slice out 4 points that lie in and outside OSGB extent.
        u = u[1:3, 3:5]
        v = v[1:3, 3:5]
        ut, vt = rotate_winds(u, v, iris.coord_systems.OSGB())
        # Values precalculated and checked.
        expected_ut_data = np.array(
            [[0.16285514, 0.35323639], [1.82650698, 2.62455840]]
        )
        expected_vt_data = np.array(
            [[19.88979966, 19.01921346], [19.88018847, 19.01424281]]
        )
        # Compare u and v data values against previously calculated values.
        self.assertArrayAllClose(ut.data, expected_ut_data, rtol=1e-5)
        self.assertArrayAllClose(vt.data, expected_vt_data, rtol=1e-5)

    def test_nd_data(self):
        u2d, y2d = self._uv_cubes_limited_extent()
        u, v = uv_cubes_3d(u2d)
        u = u[:, 1:3, 3:5]
        v = v[:, 1:3, 3:5]
        ut, vt = rotate_winds(u, v, iris.coord_systems.OSGB())
        # Values precalculated and checked (as test_data_values above),
        # then scaled by factor [1, 2, 3] along 0th dim (see uv_cubes_3d()).
        expected_ut_data = np.array(
            [[0.16285514, 0.35323639], [1.82650698, 2.62455840]]
        )
        expected_vt_data = np.array(
            [[19.88979966, 19.01921346], [19.88018847, 19.01424281]]
        )
        factor = np.array([1, 2, 3]).reshape(3, 1, 1)
        expected_ut_data = factor * expected_ut_data
        expected_vt_data = factor * expected_vt_data
        # Compare u and v data values against previously calculated values.
        self.assertArrayAlmostEqual(ut.data, expected_ut_data)
        self.assertArrayAlmostEqual(vt.data, expected_vt_data)

    def test_transposed(self):
        # Test case where the coordinates are not ordered yx in the cube.
        u, v = self._uv_cubes_limited_extent()
        # Slice out 4 points that lie in and outside OSGB extent.
        u = u[1:3, 3:5]
        v = v[1:3, 3:5]
        # Transpose cubes (in-place)
        u.transpose()
        v.transpose()
        ut, vt = rotate_winds(u, v, iris.coord_systems.OSGB())
        # Values precalculated and checked.
        expected_ut_data = np.array(
            [[0.16285514, 0.35323639], [1.82650698, 2.62455840]]
        ).T
        expected_vt_data = np.array(
            [[19.88979966, 19.01921346], [19.88018847, 19.01424281]]
        ).T
        # Compare u and v data values against previously calculated values.
        self.assertArrayAllClose(ut.data, expected_ut_data, rtol=1e-5)
        self.assertArrayAllClose(vt.data, expected_vt_data, rtol=1e-5)


class TestMasking(tests.IrisTest):
    def test_rotated_to_osgb(self):
        # Rotated Pole data with large extent.
        x = np.linspace(311.9, 391.1, 10)
        y = np.linspace(-23.6, 24.8, 8)
        u, v = uv_cubes(x, y)
        ut, vt = rotate_winds(u, v, iris.coord_systems.OSGB())

        # Ensure cells with discrepancies in magnitude are masked.
        self.assertTrue(ma.isMaskedArray(ut.data))
        self.assertTrue(ma.isMaskedArray(vt.data))

        # Snapshot of mask with fixed tolerance of atol=2e-3
        expected_mask = np.array(
            [
                [1, 1, 1, 0, 0, 0, 0, 0, 0, 1],
                [1, 1, 1, 0, 0, 0, 0, 0, 0, 1],
                [1, 1, 1, 1, 0, 0, 0, 0, 1, 1],
                [1, 1, 1, 1, 0, 0, 0, 0, 1, 1],
                [1, 1, 1, 1, 0, 0, 0, 0, 1, 1],
                [1, 1, 1, 1, 1, 0, 0, 1, 1, 1],
                [1, 1, 1, 1, 1, 0, 0, 1, 1, 1],
                [1, 1, 1, 1, 1, 0, 0, 1, 1, 1],
            ],
            np.bool,
        )
        self.assertArrayEqual(expected_mask, ut.data.mask)
        self.assertArrayEqual(expected_mask, vt.data.mask)

        # Check unmasked values have sufficiently small error in mag.
        expected_mag = np.sqrt(u.data ** 2 + v.data ** 2)
        # Use underlying data to ignore mask in calculation.
        res_mag = np.sqrt(ut.data.data ** 2 + vt.data.data ** 2)
        # Calculate percentage error (note there are no zero magnitudes
        # so we can divide safely).
        anom = 100.0 * np.abs(res_mag - expected_mag) / expected_mag
        self.assertTrue(anom[~ut.data.mask].max() < 0.1)

    def test_rotated_to_unrotated(self):
        # Suffiently accurate so that no mask is introduced.
        u, v = uv_cubes()
        ut, vt = rotate_winds(u, v, iris.coord_systems.GeogCS(6371229))
        self.assertFalse(ma.isMaskedArray(ut.data))
        self.assertFalse(ma.isMaskedArray(vt.data))


class TestRoundTrip(tests.IrisTest):
    def test_rotated_to_unrotated(self):
        # Check ability to use 2d coords as input.
        u, v = uv_cubes()
        ut, vt = rotate_winds(u, v, iris.coord_systems.GeogCS(6371229))
        # Remove  grid lat and lon, leaving 2d projection coords.
        ut.remove_coord("grid_latitude")
        vt.remove_coord("grid_latitude")
        ut.remove_coord("grid_longitude")
        vt.remove_coord("grid_longitude")
        # Change back.
        orig_cs = u.coord("grid_latitude").coord_system
        res_u, res_v = rotate_winds(ut, vt, orig_cs)
        # Check data values - limited accuracy due to numerical approx.
        self.assertArrayAlmostEqual(res_u.data, u.data, decimal=3)
        self.assertArrayAlmostEqual(res_v.data, v.data, decimal=3)
        # Check coords locations.
        x2d, y2d = np.meshgrid(
            u.coord("grid_longitude").points, u.coord("grid_latitude").points
        )
        # Shift longitude from 0 to 360 -> -180 to 180.
        x2d = np.where(x2d > 180, x2d - 360, x2d)
        res_x = res_u.coord(
            "projection_x_coordinate", coord_system=orig_cs
        ).points
        res_y = res_u.coord(
            "projection_y_coordinate", coord_system=orig_cs
        ).points
        self.assertArrayAlmostEqual(res_x, x2d)
        self.assertArrayAlmostEqual(res_y, y2d)
        res_x = res_v.coord(
            "projection_x_coordinate", coord_system=orig_cs
        ).points
        res_y = res_v.coord(
            "projection_y_coordinate", coord_system=orig_cs
        ).points
        self.assertArrayAlmostEqual(res_x, x2d)
        self.assertArrayAlmostEqual(res_y, y2d)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the function :func:`iris.analysis.maths._output_dtype`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from itertools import product
import operator

import numpy as np

from iris.analysis.maths import _output_dtype


class Test(tests.IrisTest):
    def setUp(self):
        # Operators which result in a value of the same dtype as their
        # arguments when the arguments' dtypes are the same.
        self.same_result_ops = [
            operator.add,
            operator.sub,
            operator.mul,
            operator.pow,
            operator.floordiv,
            np.add,
            np.subtract,
            np.multiply,
            np.power,
            np.floor_divide,
        ]

        self.unary_same_result_ops = [np.abs]

        # Operators which always result in a float.
        self.float_ops = [operator.truediv, np.true_divide]

        self.unary_float_ops = [np.log, np.log2, np.log10, np.exp]

        self.all_binary_ops = self.same_result_ops + self.float_ops
        self.all_unary_ops = self.unary_same_result_ops + self.unary_float_ops

        self.dtypes = [
            np.dtype("i2"),
            np.dtype("i4"),
            np.dtype("i8"),
            np.dtype("f2"),
            np.dtype("f4"),
            np.dtype("f8"),
        ]

    def _binary_error_message(
        self,
        op,
        first_dtype,
        second_dtype,
        expected_dtype,
        result_dtype,
        in_place=False,
    ):
        msg = (
            "Output for {op.__class__.__name__} {op.__name__!r} and "
            "arguments ({dt1!r}, {dt2!r}, in_place={in_place}) "
            "was {res!r}. Expected {exp!r}."
        )
        return msg.format(
            op=op,
            dt1=first_dtype,
            dt2=second_dtype,
            exp=expected_dtype,
            res=result_dtype,
            in_place=in_place,
        )

    def _unary_error_message(
        self, op, dtype, expected_dtype, result_dtype, in_place=False
    ):
        msg = (
            "Output for {op.__class__.__name__} {op.__name__!r} and "
            "arguments ({dt!r}, in_place={in_place}) was {res!r}. "
            "Expected {exp!r}."
        )
        return msg.format(
            op=op,
            dt=dtype,
            exp=expected_dtype,
            res=result_dtype,
            in_place=in_place,
        )

    def test_same_result(self):
        # Check that the result dtype is the same as the input dtypes for
        # relevant operators.
        for dtype in self.dtypes:
            for op in self.same_result_ops:
                result_dtype = _output_dtype(op, dtype, dtype)
                self.assertEqual(
                    dtype,
                    result_dtype,
                    self._binary_error_message(
                        op, dtype, dtype, dtype, result_dtype
                    ),
                )
            for op in self.unary_same_result_ops:
                result_dtype = _output_dtype(op, dtype)
                self.assertEqual(
                    dtype,
                    result_dtype,
                    self._unary_error_message(op, dtype, dtype, result_dtype),
                )

    def test_binary_float(self):
        # Check that the result dtype is a float for relevant operators.
        # Perform checks for a selection of cases.
        cases = [
            (np.dtype("i2"), np.dtype("i2"), np.dtype("f8")),
            (np.dtype("i2"), np.dtype("i4"), np.dtype("f8")),
            (np.dtype("i4"), np.dtype("i4"), np.dtype("f8")),
            (np.dtype("i2"), np.dtype("f2"), np.dtype("f4")),
            (np.dtype("i2"), np.dtype("f4"), np.dtype("f4")),
            (np.dtype("i8"), np.dtype("f2"), np.dtype("f8")),
            (np.dtype("f2"), np.dtype("f2"), np.dtype("f2")),
            (np.dtype("f4"), np.dtype("f4"), np.dtype("f4")),
            (np.dtype("f2"), np.dtype("f4"), np.dtype("f4")),
        ]
        for dtype1, dtype2, expected_dtype in cases:
            for op in self.float_ops:
                result_dtype = _output_dtype(op, dtype1, dtype2)
                self.assertEqual(
                    expected_dtype,
                    result_dtype,
                    self._binary_error_message(
                        op, dtype1, dtype2, expected_dtype, result_dtype
                    ),
                )

    def test_unary_float(self):
        cases = [
            (np.dtype("i2"), np.dtype("f4")),
            (np.dtype("i4"), np.dtype("f8")),
            (np.dtype("i8"), np.dtype("f8")),
            (np.dtype("f2"), np.dtype("f2")),
            (np.dtype("f4"), np.dtype("f4")),
            (np.dtype("f8"), np.dtype("f8")),
        ]
        for dtype, expected_dtype in cases:
            for op in self.unary_float_ops:
                result_dtype = _output_dtype(op, dtype)
                self.assertEqual(
                    expected_dtype,
                    result_dtype,
                    self._unary_error_message(
                        op, dtype, expected_dtype, result_dtype
                    ),
                )

    def test_binary_float_argument(self):
        # Check that when one argument is a float dtype, a float dtype results
        # Unary operators are covered by other tests.
        dtypes = [
            np.dtype("i2"),
            np.dtype("i4"),
            np.dtype("i8"),
            np.dtype("f2"),
            np.dtype("f4"),
            np.dtype("f8"),
        ]
        expected_dtypes = [
            np.dtype("f4"),
            np.dtype("f8"),
            np.dtype("f8"),
            np.dtype("f2"),
            np.dtype("f4"),
            np.dtype("f8"),
        ]
        for op in self.all_binary_ops:
            for dtype, expected_dtype in zip(dtypes, expected_dtypes):
                result_dtype = _output_dtype(op, dtype, np.dtype("f2"))
                self.assertEqual(
                    expected_dtype,
                    result_dtype,
                    self._binary_error_message(
                        op, dtype, np.dtype("f2"), expected_dtype, result_dtype
                    ),
                )

    def test_in_place(self):
        # Check that when the in_place argument is True, the result is always
        # the same as first operand.
        for dtype1, dtype2 in product(self.dtypes, self.dtypes):
            for op in self.all_binary_ops:
                result_dtype = _output_dtype(op, dtype1, dtype2, in_place=True)
                self.assertEqual(
                    result_dtype,
                    dtype1,
                    self._binary_error_message(
                        op, dtype1, dtype2, dtype1, result_dtype, in_place=True
                    ),
                )
        for dtype in self.dtypes:
            for op in self.all_unary_ops:
                result_dtype = _output_dtype(op, dtype, in_place=True)
                self.assertEqual(
                    result_dtype,
                    dtype,
                    self._unary_error_message(
                        op, dtype, dtype, result_dtype, in_place=True
                    ),
                )

    def test_commuative(self):
        # Check that the operation is commutative if in_place is not specified.
        for dtype1, dtype2 in product(self.dtypes, self.dtypes):
            for op in self.all_binary_ops:
                result_dtype1 = _output_dtype(op, dtype1, dtype2)
                result_dtype2 = _output_dtype(op, dtype2, dtype1)
                self.assertEqual(
                    result_dtype1,
                    result_dtype2,
                    "_output_dtype is not commutative with arguments "
                    "{!r} and {!r}: {!r} != {!r}".format(
                        dtype1, dtype2, result_dtype1, result_dtype2
                    ),
                )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
A collection of routines which create standard Cubes for test purposes.

"""

from datetime import datetime
import os.path

import numpy as np
import numpy.ma as ma

from cf_units import Unit
from iris.cube import Cube
import iris.aux_factory
import iris.coords
import iris.coords as icoords
from iris.coords import DimCoord, AuxCoord, CellMethod
import iris.tests as tests
from iris.coord_systems import GeogCS, RotatedGeogCS
from ._stock_2d_latlons import (  # noqa
    sample_2d_latlons,
    make_bounds_discontiguous_at_point,
)


def lat_lon_cube():
    """
    Returns a cube with a latitude and longitude suitable for testing
    saving to PP/NetCDF etc.

    """
    cube = Cube(np.arange(12, dtype=np.int32).reshape((3, 4)))
    cs = GeogCS(6371229)
    coord = DimCoord(
        points=np.array([-1, 0, 1], dtype=np.int32),
        standard_name="latitude",
        units="degrees",
        coord_system=cs,
    )
    cube.add_dim_coord(coord, 0)
    coord = DimCoord(
        points=np.array([-1, 0, 1, 2], dtype=np.int32),
        standard_name="longitude",
        units="degrees",
        coord_system=cs,
    )
    cube.add_dim_coord(coord, 1)
    return cube


def global_pp():
    """
    Returns a two-dimensional cube derived from PP/aPPglob1/global.pp.

    The standard_name and unit attributes are added to compensate for the
    broken STASH encoding in that file.

    """

    def callback_global_pp(cube, field, filename):
        cube.standard_name = "air_temperature"
        cube.units = "K"

    path = tests.get_data_path(("PP", "aPPglob1", "global.pp"))
    cube = iris.load_cube(path, callback=callback_global_pp)
    return cube


def simple_pp():
    # Differs from global_pp()
    filename = tests.get_data_path(["PP", "simple_pp", "global.pp"])
    cube = iris.load_cube(filename)
    return cube


def simple_1d(with_bounds=True):
    """
    Returns an abstract, one-dimensional cube.

    >>> print(simple_1d())
    thingness                           (foo: 11)
         Dimension coordinates:
              foo                           x

    >>> print(repr(simple_1d().data))
    [ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10]

    """
    cube = Cube(np.arange(11, dtype=np.int32))
    cube.long_name = "thingness"
    cube.units = "1"
    points = np.arange(11, dtype=np.int32) + 1
    bounds = np.column_stack(
        [np.arange(11, dtype=np.int32), np.arange(11, dtype=np.int32) + 1]
    )
    coord = DimCoord(points, long_name="foo", units="1", bounds=bounds)
    cube.add_dim_coord(coord, 0)
    return cube


def simple_2d(with_bounds=True):
    """
    Returns an abstract, two-dimensional, optionally bounded, cube.

    >>> print(simple_2d())
    thingness                           (bar: 3; foo: 4)
         Dimension coordinates:
              bar                           x       -
              foo                           -       x

    >>> print(repr(simple_2d().data))
    [[ 0  1  2  3]
     [ 4  5  6  7]
     [ 8  9 10 11]]


    """
    cube = Cube(np.arange(12, dtype=np.int32).reshape((3, 4)))
    cube.long_name = "thingness"
    cube.units = "1"
    y_points = np.array([2.5, 7.5, 12.5])
    y_bounds = np.array([[0, 5], [5, 10], [10, 15]], dtype=np.int32)
    y_coord = DimCoord(
        y_points,
        long_name="bar",
        units="1",
        bounds=y_bounds if with_bounds else None,
    )
    x_points = np.array([-7.5, 7.5, 22.5, 37.5])
    x_bounds = np.array(
        [[-15, 0], [0, 15], [15, 30], [30, 45]], dtype=np.int32
    )
    x_coord = DimCoord(
        x_points,
        long_name="foo",
        units="1",
        bounds=x_bounds if with_bounds else None,
    )

    cube.add_dim_coord(y_coord, 0)
    cube.add_dim_coord(x_coord, 1)
    return cube


def simple_2d_w_multidim_coords(with_bounds=True):
    """
    Returns an abstract, two-dimensional, optionally bounded, cube.

    >>> print(simple_2d_w_multidim_coords())
    thingness                           (*ANONYMOUS*: 3; *ANONYMOUS*: 4)
         Auxiliary coordinates:
              bar                                   x               x
              foo                                   x               x

    >>> print(repr(simple_2d().data))
    [[ 0,  1,  2,  3],
     [ 4,  5,  6,  7],
     [ 8,  9, 10, 11]]

    """
    cube = simple_3d_w_multidim_coords(with_bounds)[0, :, :]
    cube.remove_coord("wibble")
    cube.data = np.arange(12, dtype=np.int32).reshape((3, 4))
    return cube


def simple_3d_w_multidim_coords(with_bounds=True):
    """
    Returns an abstract, two-dimensional, optionally bounded, cube.

    >>> print(simple_3d_w_multidim_coords())
    thingness                           (wibble: 2; *ANONYMOUS*: 3; *ANONYMOUS*: 4)
         Dimension coordinates:
              wibble                           x               -               -
         Auxiliary coordinates:
              bar                              -               x               x
              foo                              -               x               x

    >>> print(simple_3d_w_multidim_coords().data)
    [[[ 0  1  2  3]
      [ 4  5  6  7]
      [ 8  9 10 11]]

     [[12 13 14 15]
      [16 17 18 19]
      [20 21 22 23]]]

    """
    cube = Cube(np.arange(24, dtype=np.int32).reshape((2, 3, 4)))
    cube.long_name = "thingness"
    cube.units = "1"

    y_points = np.array(
        [
            [2.5, 7.5, 12.5, 17.5],
            [10.0, 17.5, 27.5, 42.5],
            [15.0, 22.5, 32.5, 50.0],
        ]
    )
    y_bounds = np.array(
        [
            [[0, 5], [5, 10], [10, 15], [15, 20]],
            [[5, 15], [15, 20], [20, 35], [35, 50]],
            [[10, 20], [20, 25], [25, 40], [40, 60]],
        ],
        dtype=np.int32,
    )
    y_coord = AuxCoord(
        points=y_points,
        long_name="bar",
        units="1",
        bounds=y_bounds if with_bounds else None,
    )
    x_points = np.array(
        [
            [-7.5, 7.5, 22.5, 37.5],
            [-12.5, 4.0, 26.5, 47.5],
            [2.5, 14.0, 36.5, 44.0],
        ]
    )
    x_bounds = np.array(
        [
            [[-15, 0], [0, 15], [15, 30], [30, 45]],
            [[-25, 0], [0, 8], [8, 45], [45, 50]],
            [[-5, 10], [10, 18], [18, 55], [18, 70]],
        ],
        dtype=np.int32,
    )
    x_coord = AuxCoord(
        points=x_points,
        long_name="foo",
        units="1",
        bounds=x_bounds if with_bounds else None,
    )
    wibble_coord = DimCoord(
        np.array([10.0, 30.0], dtype=np.float32), long_name="wibble", units="1"
    )

    cube.add_dim_coord(wibble_coord, [0])
    cube.add_aux_coord(y_coord, [1, 2])
    cube.add_aux_coord(x_coord, [1, 2])
    return cube


def simple_3d():
    """
    Returns an abstract three dimensional cube.

    >>> print(simple_3d())
    thingness / (1)                     (wibble: 2; latitude: 3; longitude: 4)
     Dimension coordinates:
          wibble                           x            -             -
          latitude                         -            x             -
          longitude                        -            -             x

    >>> print(simple_3d().data)
    [[[ 0  1  2  3]
      [ 4  5  6  7]
      [ 8  9 10 11]]

     [[12 13 14 15]
      [16 17 18 19]
      [20 21 22 23]]]

    """
    cube = Cube(np.arange(24, dtype=np.int32).reshape((2, 3, 4)))
    cube.long_name = "thingness"
    cube.units = "1"
    wibble_coord = DimCoord(
        np.array([10.0, 30.0], dtype=np.float32), long_name="wibble", units="1"
    )
    lon = DimCoord(
        [-180, -90, 0, 90],
        standard_name="longitude",
        units="degrees",
        circular=True,
    )
    lat = DimCoord([90, 0, -90], standard_name="latitude", units="degrees")
    cube.add_dim_coord(wibble_coord, [0])
    cube.add_dim_coord(lat, [1])
    cube.add_dim_coord(lon, [2])
    return cube


def simple_3d_mask():
    """
    Returns an abstract three dimensional cube that has data masked.

    >>> print(simple_3d_mask())
    thingness / (1)                     (wibble: 2; latitude: 3; longitude: 4)
     Dimension coordinates:
          wibble                           x            -             -
          latitude                         -            x             -
          longitude                        -            -             x

    >>> print(simple_3d_mask().data)
    [[[-- -- -- --]
      [-- -- -- --]
      [-- 9 10 11]]

    [[12 13 14 15]
     [16 17 18 19]
     [20 21 22 23]]]

    """
    cube = simple_3d()
    cube.data = ma.asanyarray(cube.data)
    cube.data = ma.masked_less_equal(cube.data, 8.0)
    return cube


def track_1d(duplicate_x=False):
    """
    Returns a one-dimensional track through two-dimensional space.

    >>> print(track_1d())
    air_temperature                     (y, x: 11)
         Dimensioned coords:
              x -> x
              y -> y
         Single valued coords:

    >>> print(repr(track_1d().data))
    array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10])

    """
    cube = Cube(
        np.arange(11, dtype=np.int32),
        standard_name="air_temperature",
        units="K",
    )
    bounds = np.column_stack(
        [np.arange(11, dtype=np.int32), np.arange(11, dtype=np.int32) + 1]
    )
    pts = bounds[:, 1]
    coord = AuxCoord(pts, "projection_x_coordinate", units="1", bounds=bounds)
    cube.add_aux_coord(coord, [0])
    if duplicate_x:
        coord = AuxCoord(
            pts, "projection_x_coordinate", units="1", bounds=bounds
        )
        cube.add_aux_coord(coord, [0])
    coord = AuxCoord(
        pts * 2, "projection_y_coordinate", units="1", bounds=bounds * 2
    )
    cube.add_aux_coord(coord, 0)
    return cube


def simple_2d_w_multidim_and_scalars():
    data = np.arange(50, dtype=np.int32).reshape((5, 10))
    cube = iris.cube.Cube(
        data, long_name="test 2d dimensional cube", units="meters"
    )

    # DimCoords
    dim1 = DimCoord(
        np.arange(5, dtype=np.float32) * 5.1 + 3.0,
        long_name="dim1",
        units="meters",
    )
    dim2 = DimCoord(
        np.arange(10, dtype=np.int32),
        long_name="dim2",
        units="meters",
        bounds=np.arange(20, dtype=np.int32).reshape(10, 2),
    )

    # Scalars
    an_other = AuxCoord(3.0, long_name="an_other", units="meters")
    yet_an_other = DimCoord(
        23.3,
        standard_name="air_temperature",
        long_name="custom long name",
        var_name="custom_var_name",
        units="K",
    )

    # Multidim
    my_multi_dim_coord = AuxCoord(
        np.arange(50, dtype=np.int32).reshape(5, 10),
        long_name="my_multi_dim_coord",
        units="1",
        bounds=np.arange(200, dtype=np.int32).reshape(5, 10, 4),
    )

    cube.add_dim_coord(dim1, 0)
    cube.add_dim_coord(dim2, 1)
    cube.add_aux_coord(an_other)
    cube.add_aux_coord(yet_an_other)
    cube.add_aux_coord(my_multi_dim_coord, [0, 1])

    return cube


def hybrid_height():
    """
    Returns a two-dimensional (Z, X), hybrid-height cube.

    >>> print(hybrid_height())
    TODO: Update!
    air_temperature                     (level_height: 3; *ANONYMOUS*: 4)
         Dimension coordinates:
              level_height                           x               -
         Auxiliary coordinates:
              model_level_number                     x               -
              sigma                                  x               -
              surface_altitude                       -               x
         Derived coordinates:
              altitude                               x               x

    >>> print(hybrid_height().data)
    [[[ 0  1  2  3]
      [ 4  5  6  7]
      [ 8  9 10 11]]

    """
    data = np.arange(12, dtype="i8").reshape((3, 4))

    orography = AuxCoord(
        [10, 25, 50, 5], standard_name="surface_altitude", units="m"
    )
    model_level = AuxCoord([2, 1, 0], standard_name="model_level_number")
    level_height = DimCoord(
        [100, 50, 10],
        long_name="level_height",
        units="m",
        attributes={"positive": "up"},
        bounds=[[150, 75], [75, 20], [20, 0]],
    )
    sigma = AuxCoord(
        [0.8, 0.9, 0.95],
        long_name="sigma",
        bounds=[[0.7, 0.85], [0.85, 0.97], [0.97, 1.0]],
    )
    hybrid_height = iris.aux_factory.HybridHeightFactory(
        level_height, sigma, orography
    )

    cube = iris.cube.Cube(
        data,
        standard_name="air_temperature",
        units="K",
        dim_coords_and_dims=[(level_height, 0)],
        aux_coords_and_dims=[(orography, 1), (model_level, 0), (sigma, 0)],
        aux_factories=[hybrid_height],
    )
    return cube


def simple_4d_with_hybrid_height():
    cube = iris.cube.Cube(
        np.arange(3 * 4 * 5 * 6, dtype="i8").reshape(3, 4, 5, 6),
        "air_temperature",
        units="K",
    )

    cube.add_dim_coord(
        DimCoord(np.arange(3, dtype="i8"), "time", units="hours since epoch"),
        0,
    )
    cube.add_dim_coord(
        DimCoord(
            np.arange(4, dtype="i8") + 10, "model_level_number", units="1"
        ),
        1,
    )
    cube.add_dim_coord(
        DimCoord(
            np.arange(5, dtype="i8") + 20, "grid_latitude", units="degrees"
        ),
        2,
    )
    cube.add_dim_coord(
        DimCoord(
            np.arange(6, dtype="i8") + 30, "grid_longitude", units="degrees"
        ),
        3,
    )

    cube.add_aux_coord(
        AuxCoord(
            np.arange(4, dtype="i8") + 40, long_name="level_height", units="m"
        ),
        1,
    )
    cube.add_aux_coord(
        AuxCoord(np.arange(4, dtype="i8") + 50, long_name="sigma", units="1"),
        1,
    )
    cube.add_aux_coord(
        AuxCoord(
            np.arange(5 * 6, dtype="i8").reshape(5, 6) + 100,
            long_name="surface_altitude",
            units="m",
        ),
        [2, 3],
    )

    cube.add_aux_factory(
        iris.aux_factory.HybridHeightFactory(
            delta=cube.coord("level_height"),
            sigma=cube.coord("sigma"),
            orography=cube.coord("surface_altitude"),
        )
    )
    return cube


def realistic_3d():
    """
    Returns a realistic 3d cube.

    >>> print(repr(realistic_3d()))
    <iris 'Cube' of air_potential_temperature (time: 7; grid_latitude: 9;
    grid_longitude: 11)>

    """
    data = np.arange(7 * 9 * 11).reshape((7, 9, 11))
    lat_pts = np.linspace(-4, 4, 9)
    lon_pts = np.linspace(-5, 5, 11)
    time_pts = np.linspace(394200, 394236, 7)
    forecast_period_pts = np.linspace(0, 36, 7)
    ll_cs = RotatedGeogCS(37.5, 177.5, ellipsoid=GeogCS(6371229.0))

    lat = icoords.DimCoord(
        lat_pts,
        standard_name="grid_latitude",
        units="degrees",
        coord_system=ll_cs,
    )
    lon = icoords.DimCoord(
        lon_pts,
        standard_name="grid_longitude",
        units="degrees",
        coord_system=ll_cs,
    )
    time = icoords.DimCoord(
        time_pts, standard_name="time", units="hours since 1970-01-01 00:00:00"
    )
    forecast_period = icoords.DimCoord(
        forecast_period_pts, standard_name="forecast_period", units="hours"
    )
    height = icoords.DimCoord(1000.0, standard_name="air_pressure", units="Pa")
    cube = iris.cube.Cube(
        data,
        standard_name="air_potential_temperature",
        units="K",
        dim_coords_and_dims=[(time, 0), (lat, 1), (lon, 2)],
        aux_coords_and_dims=[(forecast_period, 0), (height, None)],
        attributes={"source": "Iris test case"},
    )
    return cube


def realistic_4d():
    """
    Returns a realistic 4d cube.

    >>> print(repr(realistic_4d()))
    <iris 'Cube' of air_potential_temperature (time: 6; model_level_number: 70;
    grid_latitude: 100; grid_longitude: 100)>

    """
    data_path = tests.get_data_path(("stock", "stock_arrays.npz"))
    if not os.path.isfile(data_path):
        raise IOError("Test data is not available at {}.".format(data_path))
    r = np.load(data_path)
    # sort the arrays based on the order they were originally given.
    # The names given are of the form 'arr_1' or 'arr_10'
    _, arrays = zip(*sorted(r.items(), key=lambda item: int(item[0][4:])))

    (
        lat_pts,
        lat_bnds,
        lon_pts,
        lon_bnds,
        level_height_pts,
        level_height_bnds,
        model_level_pts,
        sigma_pts,
        sigma_bnds,
        time_pts,
        _source_pts,
        forecast_period_pts,
        data,
        orography,
    ) = arrays

    ll_cs = RotatedGeogCS(37.5, 177.5, ellipsoid=GeogCS(6371229.0))

    lat = icoords.DimCoord(
        lat_pts,
        standard_name="grid_latitude",
        units="degrees",
        bounds=lat_bnds,
        coord_system=ll_cs,
    )
    lon = icoords.DimCoord(
        lon_pts,
        standard_name="grid_longitude",
        units="degrees",
        bounds=lon_bnds,
        coord_system=ll_cs,
    )
    level_height = icoords.DimCoord(
        level_height_pts,
        long_name="level_height",
        units="m",
        bounds=level_height_bnds,
        attributes={"positive": "up"},
    )
    model_level = icoords.DimCoord(
        model_level_pts,
        standard_name="model_level_number",
        units="1",
        attributes={"positive": "up"},
    )
    sigma = icoords.AuxCoord(
        sigma_pts, long_name="sigma", units="1", bounds=sigma_bnds
    )
    orography = icoords.AuxCoord(
        orography, standard_name="surface_altitude", units="m"
    )
    time = icoords.DimCoord(
        time_pts, standard_name="time", units="hours since 1970-01-01 00:00:00"
    )
    forecast_period = icoords.DimCoord(
        forecast_period_pts, standard_name="forecast_period", units="hours"
    )

    hybrid_height = iris.aux_factory.HybridHeightFactory(
        level_height, sigma, orography
    )

    cube = iris.cube.Cube(
        data,
        standard_name="air_potential_temperature",
        units="K",
        dim_coords_and_dims=[(time, 0), (model_level, 1), (lat, 2), (lon, 3)],
        aux_coords_and_dims=[
            (orography, (2, 3)),
            (level_height, 1),
            (sigma, 1),
            (forecast_period, None),
        ],
        attributes={"source": "Iris test case"},
        aux_factories=[hybrid_height],
    )
    return cube


def realistic_4d_no_derived():
    """
    Returns a realistic 4d cube without hybrid height

    >>> print(repr(realistic_4d()))
    <iris 'Cube' of air_potential_temperature (time: 6; model_level_number: 70;
    grid_latitude: 100; grid_longitude: 100)>

    """
    cube = realistic_4d()

    # TODO determine appropriate way to remove aux_factory from a cube
    cube._aux_factories = []

    return cube


def realistic_4d_w_missing_data():
    data_path = tests.get_data_path(("stock", "stock_mdi_arrays.npz"))
    if not os.path.isfile(data_path):
        raise IOError("Test data is not available at {}.".format(data_path))
    data_archive = np.load(data_path)
    data = ma.masked_array(data_archive["arr_0"], mask=data_archive["arr_1"])

    # sort the arrays based on the order they were originally given.
    # The names given are of the form 'arr_1' or 'arr_10'

    ll_cs = GeogCS(6371229)

    lat = DimCoord(
        np.arange(20, dtype=np.float32),
        standard_name="grid_latitude",
        units="degrees",
        coord_system=ll_cs,
    )
    lon = DimCoord(
        np.arange(20, dtype=np.float32),
        standard_name="grid_longitude",
        units="degrees",
        coord_system=ll_cs,
    )
    time = DimCoord(
        [1000.0, 1003.0, 1006.0],
        standard_name="time",
        units="hours since 1970-01-01 00:00:00",
    )
    forecast_period = DimCoord(
        [0.0, 3.0, 6.0], standard_name="forecast_period", units="hours"
    )
    pressure = DimCoord(
        np.array([800.0, 900.0, 1000.0], dtype=np.float32),
        long_name="pressure",
        units="hPa",
    )

    cube = iris.cube.Cube(
        data,
        long_name="missing data test data",
        units="K",
        dim_coords_and_dims=[(time, 0), (pressure, 1), (lat, 2), (lon, 3)],
        aux_coords_and_dims=[(forecast_period, 0)],
        attributes={"source": "Iris test case"},
    )
    return cube


def ocean_sigma_z():
    """
    Return a sample cube with an
    :class:`iris.aux_factory.OceanSigmaZFactory` vertical coordinate.

    This is a fairly small cube with real coordinate arrays.  The coordinate
    values are derived from the sample data linked at
    https://github.com/SciTools/iris/pull/509#issuecomment-23565381.

    """
    co_time = DimCoord([0.0, 1.0], standard_name="time", units="")
    co_lats = DimCoord(
        [-58.1, -52.7, -46.9], standard_name="latitude", units=Unit("degrees")
    )
    co_lons = DimCoord(
        [65.1, 72.9, 83.7, 96.5],
        standard_name="longitude",
        units=Unit("degrees"),
    )
    co_ssh = AuxCoord(
        [
            [
                [-0.63157895, -0.52631579, -0.42105263, -0.31578947],
                [-0.78947368, -0.68421053, -0.57894737, -0.47368421],
                [-0.94736842, -0.84210526, -0.73684211, -0.63157895],
            ],
            [
                [-0.84210526, -0.73684211, -0.63157895, -0.52631579],
                [-1.00000000, -0.89473684, -0.78947368, -0.68421053],
                [-1.15789474, -1.05263158, -0.94736842, -0.84210526],
            ],
        ],
        standard_name="sea_surface_height",
        units=Unit("m"),
    )

    co_sigma = AuxCoord(
        [0.0, -0.1, -0.6, -1.0, -1.0],
        standard_name="ocean_sigma_z_coordinate",
        units=Unit("1"),
        attributes={"positive": "up"},
    )

    co_zlay = AuxCoord(
        [-137.2, -137.3, -137.4, -368.4, -1495.6],
        long_name="layer_depth",
        units=Unit("m"),
    )
    co_depth = AuxCoord(
        [
            [1625.7, 3921.2, 4106.4, 5243.5],
            [3615.4, 4942.6, 3883.6, 4823.1],
            [3263.2, 2816.3, 2741.8, 3883.6],
        ],
        standard_name="depth",
        units=Unit("m"),
    )
    co_depthc = DimCoord(137.9, long_name="depth_c", units=Unit("m"))
    co_nsigma = DimCoord(3, long_name="nsigma")

    cube = Cube(np.zeros((2, 5, 3, 4)))
    cube.add_dim_coord(co_time, 0)
    cube.add_dim_coord(co_lats, 2)
    cube.add_dim_coord(co_lons, 3)
    cube.add_aux_coord(co_zlay, 1)
    cube.add_aux_coord(co_sigma, 1)
    cube.add_aux_coord(co_ssh, (0, 2, 3))
    cube.add_aux_coord(co_depth, (2, 3))
    cube.add_aux_coord(co_depthc)
    cube.add_aux_coord(co_nsigma)

    fact = iris.aux_factory.OceanSigmaZFactory(
        depth=co_depth,
        eta=co_ssh,
        depth_c=co_depthc,
        zlev=co_zlay,
        sigma=co_sigma,
        nsigma=co_nsigma,
    )
    cube.add_aux_factory(fact)
    return cube


def climatology_3d():
    def jan_offset(day, year):
        dt = datetime(year, 1, day) - datetime(1970, 1, 1)
        return dt.total_seconds() / (24.0 * 3600)

    days = range(10, 15)
    years = [[year, year + 10] for year in [2001] * 4]
    days_since = [
        [jan_offset(day, yr1), jan_offset(day, yr2)]
        for (day, [yr1, yr2]) in zip(days, years)
    ]
    time_bounds = np.array(days_since)
    time_points = time_bounds[..., 0]

    lon = np.linspace(-25, 25, 5)
    lat = np.linspace(0, 60, 3)

    time_dim = DimCoord(
        time_points,
        standard_name="time",
        bounds=time_bounds,
        units="days since 1970-01-01 00:00:00-00",
        climatological=True,
    )
    lon_dim = DimCoord(lon, standard_name="longitude", units="degrees")
    lat_dim = DimCoord(lat, standard_name="latitude", units="degrees")

    data_shape = (len(time_points), len(lat), len(lon))
    values = np.zeros(shape=data_shape, dtype=np.int8)
    cube = Cube(values)
    cube.add_dim_coord(time_dim, 0)
    cube.add_dim_coord(lat_dim, 1)
    cube.add_dim_coord(lon_dim, 2)
    cube.rename("climatology test")
    cube.units = "Kelvin"
    cube.add_cell_method(CellMethod("mean over years", coords="time"))

    return cube



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Tests for NAME loading."""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests
import iris


@tests.skip_data
class TestLoad(tests.IrisTest):
    def test_NAMEIII_field(self):
        cubes = iris.load(tests.get_data_path(("NAME", "NAMEIII_field.txt")))
        self.assertCMLApproxData(cubes, ("name", "NAMEIII_field.cml"))

    def test_NAMEII_field(self):
        cubes = iris.load(tests.get_data_path(("NAME", "NAMEII_field.txt")))
        self.assertCMLApproxData(cubes, ("name", "NAMEII_field.cml"))

    def test_NAMEIII_timeseries(self):
        cubes = iris.load(
            tests.get_data_path(("NAME", "NAMEIII_timeseries.txt"))
        )
        self.assertCMLApproxData(cubes, ("name", "NAMEIII_timeseries.cml"))

    def test_NAMEII_timeseries(self):
        cubes = iris.load(
            tests.get_data_path(("NAME", "NAMEII_timeseries.txt"))
        )
        self.assertCMLApproxData(cubes, ("name", "NAMEII_timeseries.cml"))

    def test_NAMEIII_version2(self):
        cubes = iris.load(
            tests.get_data_path(("NAME", "NAMEIII_version2.txt"))
        )
        self.assertCMLApproxData(cubes, ("name", "NAMEIII_version2.cml"))

    def test_NAMEII_trajectory(self):
        cubes = iris.load(
            tests.get_data_path(("NAME", "NAMEIII_trajectory.txt"))
        )
        self.assertCML(cubes[0], ("name", "NAMEIII_trajectory0.cml"))
        self.assertCML(
            cubes, ("name", "NAMEIII_trajectory.cml"), checksum=False
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.fileformats.pp.save_fields` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import numpy as np

import iris.fileformats.pp as pp


def asave(afilehandle):
    afilehandle.write("saved")


class TestSaveFields(tests.IrisTest):
    def setUp(self):
        # Create a test object to stand in for a real PPField.
        self.pp_field = mock.MagicMock(spec=pp.PPField3)
        # Add minimal content required by the pp.save operation.
        self.pp_field.HEADER_DEFN = pp.PPField3.HEADER_DEFN
        self.pp_field.data = np.zeros((1, 1))
        self.pp_field.save = asave

    def test_save(self):
        open_func = "builtins.open"
        m = mock.mock_open()
        with mock.patch(open_func, m, create=True):
            pp.save_fields([self.pp_field], "foo.pp")
        self.assertTrue(mock.call("foo.pp", "wb") in m.mock_calls)
        self.assertTrue(mock.call().write("saved") in m.mock_calls)

    def test_save_append(self):
        open_func = "builtins.open"
        m = mock.mock_open()
        with mock.patch(open_func, m, create=True):
            pp.save_fields([self.pp_field], "foo.pp", append=True)
        self.assertTrue(mock.call("foo.pp", "ab") in m.mock_calls)
        self.assertTrue(mock.call().write("saved") in m.mock_calls)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris.fileformats._pyke_rules.compiled_krb.\
fc_rules_cf_fc.build_cube_metadata`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from unittest import mock

import numpy as np

from iris.cube import Cube
from iris.fileformats._pyke_rules.compiled_krb.fc_rules_cf_fc import \
    build_cube_metadata


def _make_engine(global_attributes=None, standard_name=None, long_name=None):
    if global_attributes is None:
        global_attributes = {}

    cf_group = mock.Mock(global_attributes=global_attributes)

    cf_var = mock.MagicMock(
        cf_name='wibble',
        standard_name=standard_name,
        long_name=long_name,
        units='m',
        dtype=np.float64,
        cell_methods=None,
        cf_group=cf_group)

    engine = mock.Mock(
        cube=Cube([23]),
        cf_var=cf_var)

    return engine


class TestInvalidGlobalAttributes(tests.IrisTest):
    def test_valid(self):
        global_attributes = {'Conventions': 'CF-1.5',
                             'comment': 'Mocked test object'}
        engine = _make_engine(global_attributes)
        build_cube_metadata(engine)
        expected = global_attributes
        self.assertEqual(engine.cube.attributes, expected)

    def test_invalid(self):
        global_attributes = {'Conventions': 'CF-1.5',
                             'comment': 'Mocked test object',
                             'calendar': 'standard'}
        engine = _make_engine(global_attributes)
        with mock.patch('warnings.warn') as warn:
            build_cube_metadata(engine)
        # Check for a warning.
        self.assertEqual(warn.call_count, 1)
        self.assertIn("Skipping global attribute 'calendar'",
                      warn.call_args[0][0])
        # Check resulting attributes. The invalid entry 'calendar'
        # should be filtered out.
        global_attributes.pop('calendar')
        expected = global_attributes
        self.assertEqual(engine.cube.attributes, expected)


class TestCubeName(tests.IrisTest):
    def check_cube_names(self, inputs, expected):
        # Inputs - attributes on the fake CF Variable.
        standard_name, long_name = inputs
        # Expected - The expected cube attributes.
        exp_standard_name, exp_long_name = expected

        engine = _make_engine(standard_name=standard_name, long_name=long_name)
        build_cube_metadata(engine)

        # Check the cube's standard name and long name are as expected.
        self.assertEqual(engine.cube.standard_name, exp_standard_name)
        self.assertEqual(engine.cube.long_name, exp_long_name)

    def test_standard_name_none_long_name_none(self):
        inputs = (None, None)
        expected = (None, None)
        self.check_cube_names(inputs, expected)

    def test_standard_name_none_long_name_set(self):
        inputs = (None, 'ice_thickness_long_name')
        expected = (None, 'ice_thickness_long_name')
        self.check_cube_names(inputs, expected)

    def test_standard_name_valid_long_name_none(self):
        inputs = ('sea_ice_thickness', None)
        expected = ('sea_ice_thickness', None)
        self.check_cube_names(inputs, expected)

    def test_standard_name_valid_long_name_set(self):
        inputs = ('sea_ice_thickness', 'ice_thickness_long_name')
        expected = ('sea_ice_thickness', 'ice_thickness_long_name')
        self.check_cube_names(inputs, expected)

    def test_standard_name_invalid_long_name_none(self):
        inputs = ('not_a_standard_name', None)
        expected = (None, 'not_a_standard_name',)
        self.check_cube_names(inputs, expected)

    def test_standard_name_invalid_long_name_set(self):
        inputs = ('not_a_standard_name', 'ice_thickness_long_name')
        expected = (None, 'ice_thickness_long_name')
        self.check_cube_names(inputs, expected)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the function :func:`iris.analysis.maths._get_dtype`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np
from numpy import ma

from iris.analysis.maths import _get_dtype
from iris.cube import Cube
from iris.coords import DimCoord, AuxCoord


class Test(tests.IrisTest):
    def _check_call(self, obj, expected_dtype):
        result = _get_dtype(obj)
        self.assertEqual(expected_dtype, result)

    def test_int8(self):
        n = -128
        self._check_call(n, np.int8)

    def test_int16(self):
        n = -129
        self._check_call(n, np.int16)

    def test_uint8(self):
        n = 255
        self._check_call(n, np.uint8)

    def test_uint16(self):
        n = 256
        self._check_call(n, np.uint16)

    def test_float16(self):
        n = 60000.0
        self._check_call(n, np.float16)

    def test_float32(self):
        n = 65000.0
        self._check_call(n, np.float32)

    def test_float64(self):
        n = 1e40
        self._check_call(n, np.float64)

    def test_scalar_demote(self):
        n = np.int64(10)
        self._check_call(n, np.uint8)

    def test_array(self):
        a = np.array([1, 2, 3], dtype=np.int16)
        self._check_call(a, np.int16)

    def test_scalar_array(self):
        dtype = np.int32
        a = np.array(1, dtype=dtype)
        self._check_call(a, dtype)

    def test_masked_array(self):
        dtype = np.float16
        m = ma.masked_array([1, 2, 3], [1, 0, 1], dtype=dtype)
        self._check_call(m, dtype)

    def test_masked_constant(self):
        m = ma.masked
        self._check_call(m, m.dtype)

    def test_cube(self):
        dtype = np.float32
        data = np.array([1, 2, 3], dtype=dtype)
        cube = Cube(data)
        self._check_call(cube, dtype)

    def test_aux_coord(self):
        dtype = np.int64
        points = np.array([1, 2, 3], dtype=dtype)
        aux_coord = AuxCoord(points)
        self._check_call(aux_coord, dtype)

    def test_dim_coord(self):
        dtype = np.float16
        points = np.array([1, 2, 3], dtype=dtype)
        dim_coord = DimCoord(points)
        self._check_call(dim_coord, dtype)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

import contextlib
import os.path

import iris


class PPTest:
    """
    A mixin class to provide PP-specific utilities to subclasses of tests.IrisTest.

    """

    @contextlib.contextmanager
    def cube_save_test(
        self,
        reference_txt_path,
        reference_cubes=None,
        reference_pp_path=None,
        **kwargs,
    ):
        """
        A context manager for testing the saving of Cubes to PP files.

        Args:

        * reference_txt_path:
            The path of the file containing the textual PP reference data.

        Kwargs:

        * reference_cubes:
            The cube(s) from which the textual PP reference can be re-built if necessary.
        * reference_pp_path:
            The location of a PP file from which the textual PP reference can be re-built if necessary.
            NB. The "reference_cubes" argument takes precedence over this argument.

        The return value from the context manager is the name of a temporary file
        into which the PP data to be tested should be saved.

        Example::
            with self.cube_save_test(reference_txt_path, reference_cubes=cubes) as temp_pp_path:
                iris.save(cubes, temp_pp_path)

        """
        # Watch out for a missing reference text file
        if not os.path.isfile(reference_txt_path):
            if reference_cubes:
                temp_pp_path = iris.util.create_temp_filename(".pp")
                try:
                    iris.save(reference_cubes, temp_pp_path, **kwargs)
                    self._create_reference_txt(
                        reference_txt_path, temp_pp_path
                    )
                finally:
                    os.remove(temp_pp_path)
            elif reference_pp_path:
                self._create_reference_txt(
                    reference_txt_path, reference_pp_path
                )
            else:
                raise ValueError(
                    "Missing all of reference txt file, cubes, and PP path."
                )

        temp_pp_path = iris.util.create_temp_filename(".pp")
        try:
            # This value is returned to the target of the "with" statement's "as" clause.
            yield temp_pp_path

            # Load deferred data for all of the fields (but don't do anything with it)
            pp_fields = list(iris.fileformats.pp.load(temp_pp_path))
            for pp_field in pp_fields:
                pp_field.data
            with open(reference_txt_path, "r") as reference_fh:
                reference = "".join(reference_fh)
            self._assert_str_same(
                reference + "\n",
                str(pp_fields) + "\n",
                reference_txt_path,
                type_comparison_name="PP files",
            )
        finally:
            os.remove(temp_pp_path)

    def _create_reference_txt(self, txt_path, pp_path):
        # Load the reference data
        pp_fields = list(iris.fileformats.pp.load(pp_path))
        for pp_field in pp_fields:
            pp_field.data

        # Clear any header words we don't use
        unused = ("lbexp", "lbegin", "lbnrec", "lbproj", "lbtyp")
        for pp_field in pp_fields:
            for word_name in unused:
                setattr(pp_field, word_name, 0)

        # Save the textual representation of the PP fields
        with open(txt_path, "w") as txt_file:
            txt_file.writelines(str(pp_fields))



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :mod:`iris.analysis.maths` module."""

from abc import ABCMeta, abstractmethod

import numpy as np
from numpy import ma

from iris.analysis import MEAN
from iris.coords import DimCoord
from iris.cube import Cube
import iris.tests as tests
import iris.tests.stock as stock


class CubeArithmeticBroadcastingTestMixin(metaclass=ABCMeta):
    # A framework for testing the broadcasting behaviour of the various cube
    # arithmetic operations.  (A test for each operation inherits this).
    @property
    @abstractmethod
    def data_op(self):
        # Define an operator to be called, I.E. 'operator.xx'.
        pass

    @property
    @abstractmethod
    def cube_func(self):
        # Define an iris arithmetic function to be called
        # I.E. 'iris.analysis.maths.xx'.
        pass

    def test_transposed(self):
        cube = stock.realistic_4d_no_derived()
        other = cube.copy()
        other.transpose()
        res = self.cube_func(cube, other)
        self.assertCML(res, checksum=False)
        expected_data = self.data_op(cube.data, other.data.T)
        self.assertArrayEqual(res.data, expected_data)

    def test_collapse_zeroth_dim(self):
        cube = stock.realistic_4d_no_derived()
        other = cube.collapsed("time", MEAN)
        res = self.cube_func(cube, other)
        self.assertCML(res, checksum=False)
        # No modification to other.data is needed as numpy broadcasting
        # should be sufficient.
        expected_data = self.data_op(cube.data, other.data)
        # Use assertMaskedArrayEqual as collapsing with MEAN results
        # in a cube with a masked data array.
        self.assertMaskedArrayEqual(res.data, expected_data)

    def test_collapse_all_dims(self):
        cube = stock.realistic_4d_no_derived()
        other = cube.collapsed(cube.coords(dim_coords=True), MEAN)
        res = self.cube_func(cube, other)
        self.assertCML(res, checksum=False)
        # No modification to other.data is needed as numpy broadcasting
        # should be sufficient.
        expected_data = self.data_op(cube.data, other.data)
        # Use assertArrayEqual rather than assertMaskedArrayEqual as
        # collapsing all dims does not result in a masked array.
        self.assertArrayEqual(res.data, expected_data)

    def test_collapse_last_dims(self):
        cube = stock.realistic_4d_no_derived()
        other = cube.collapsed(["grid_latitude", "grid_longitude"], MEAN)
        res = self.cube_func(cube, other)
        self.assertCML(res, checksum=False)
        # Transpose the dimensions in self.cube that have been collapsed in
        # other to lie at the front, thereby enabling numpy broadcasting to
        # function when applying data operator. Finish by transposing back
        # again to restore order.
        expected_data = self.data_op(
            cube.data.transpose((2, 3, 0, 1)), other.data
        ).transpose(2, 3, 0, 1)
        self.assertMaskedArrayEqual(res.data, expected_data)

    def test_collapse_middle_dim(self):
        cube = stock.realistic_4d_no_derived()
        other = cube.collapsed(["model_level_number"], MEAN)
        res = self.cube_func(cube, other)
        self.assertCML(res, checksum=False)
        # Add the collapsed dimension back in via np.newaxis to enable
        # numpy broadcasting to function.
        expected_data = self.data_op(cube.data, other.data[:, np.newaxis, ...])
        self.assertMaskedArrayEqual(res.data, expected_data)

    def test_slice(self):
        cube = stock.realistic_4d_no_derived()
        for dim in range(cube.ndim):
            keys = [slice(None)] * cube.ndim
            keys[dim] = 3
            other = cube[tuple(keys)]
            res = self.cube_func(cube, other)
            self.assertCML(res, checksum=False)
            # Add the collapsed dimension back in via np.newaxis to enable
            # numpy broadcasting to function.
            keys[dim] = np.newaxis
            expected_data = self.data_op(cube.data, other.data[tuple(keys)])
            msg = "Problem broadcasting cubes when sliced on dimension {}."
            self.assertArrayEqual(
                res.data, expected_data, err_msg=msg.format(dim)
            )


class CubeArithmeticMaskingTestMixin(metaclass=ABCMeta):
    # A framework for testing the mask handling behaviour of the various cube
    # arithmetic operations.  (A test for each operation inherits this).
    @property
    @abstractmethod
    def data_op(self):
        # Define an operator to be called, I.E. 'operator.xx'.
        pass

    @property
    @abstractmethod
    def cube_func(self):
        # Define an iris arithmetic function to be called
        # I.E. 'iris.analysis.maths.xx'.
        pass

    def _test_partial_mask(self, in_place):
        # Helper method for masked data tests.
        dat_a = ma.array([2.0, 2.0, 2.0, 2.0], mask=[1, 0, 1, 0])
        dat_b = ma.array([2.0, 2.0, 2.0, 2.0], mask=[1, 1, 0, 0])

        cube_a = Cube(dat_a)
        cube_b = Cube(dat_b)

        com = self.data_op(dat_b, dat_a)
        res = self.cube_func(cube_b, cube_a, in_place=in_place)

        return com, res, cube_b

    def test_partial_mask_in_place(self):
        # Cube in_place arithmetic operation.
        com, res, orig_cube = self._test_partial_mask(True)

        self.assertMaskedArrayEqual(com, res.data, strict=True)
        self.assertIs(res, orig_cube)

    def test_partial_mask_not_in_place(self):
        # Cube arithmetic not an in_place operation.
        com, res, orig_cube = self._test_partial_mask(False)

        self.assertMaskedArrayEqual(com, res.data)
        self.assertIsNot(res, orig_cube)


class CubeArithmeticCoordsTest(tests.IrisTest):
    # This class sets up pairs of cubes to test iris' ability to reject
    # arithmetic operations on coordinates which do not match.
    def SetUpNonMatching(self):
        # On this cube pair, the coordinates to perform operations on do not
        # match in either points array or name.
        data = np.zeros((3, 4))
        a = DimCoord([1, 2, 3], long_name="a")
        b = DimCoord([1, 2, 3, 4], long_name="b")
        x = DimCoord([4, 5, 6], long_name="x")
        y = DimCoord([5, 6, 7, 8], long_name="y")

        nomatch1 = Cube(data, dim_coords_and_dims=[(a, 0), (b, 1)])
        nomatch2 = Cube(data, dim_coords_and_dims=[(x, 0), (y, 1)])

        return nomatch1, nomatch2

    def SetUpReversed(self):
        # On this cube pair, the coordinates to perform operations on have
        # matching long names but the points array on one cube is reversed
        # with respect to that on the other.
        data = np.zeros((3, 4))
        a1 = DimCoord([1, 2, 3], long_name="a")
        b1 = DimCoord([1, 2, 3, 4], long_name="b")
        a2 = DimCoord([3, 2, 1], long_name="a")
        b2 = DimCoord([1, 2, 3, 4], long_name="b")

        reversed1 = Cube(data, dim_coords_and_dims=[(a1, 0), (b1, 1)])
        reversed2 = Cube(data, dim_coords_and_dims=[(a2, 0), (b2, 1)])

        return reversed1, reversed2


class CubeArithmeticMaskedConstantTestMixin(metaclass=ABCMeta):
    @property
    @abstractmethod
    def cube_func(self):
        # Define an iris arithmetic function to be called
        # I.E. 'iris.analysis.maths.xx'.
        pass

    def test_masked_constant_in_place(self):
        # Cube in_place arithmetic operation.
        dtype = np.int64
        dat = ma.masked_array(0, 1, dtype)
        cube = Cube(dat)
        res = self.cube_func(cube, 5, in_place=True)
        self.assertMaskedArrayEqual(ma.masked_array(0, 1), res.data)
        self.assertEqual(dtype, res.dtype)
        self.assertIs(res, cube)

    def test_masked_constant_not_in_place(self):
        # Cube in_place arithmetic operation.
        dtype = np.int64
        dat = ma.masked_array(0, 1, dtype)
        cube = Cube(dat)
        res = self.cube_func(cube, 5, in_place=False)
        self.assertMaskedArrayEqual(ma.masked_array(0, 1), res.data)
        self.assertEqual(dtype, res.dtype)
        self.assertIsNot(res, cube)
